"""
`hooks` module
--------------



"""

import math as m


def durkhop(D: float, ra: float = 0.3) -> object:
    """This function generates p-multipliers as per Durkhop et al (see [Duhr09]_) for application
    to cyclic curves of the `openpile.soilmodels.API_sand` model.

    The default behaviour, i.e. when the parameter `ra` = 0.3 is that the API sand cyclic curves are unchanged
    with a equivalent number of cycles equal to 100.
    When the parameter `ra` is equal to 1.0, the cyclic curves are approach the static curves.
    In between 0.3 and 1.0, the multipliers may be thought as cyclic curves with a lower number of equivalent cycles.

    Parameters
    ----------
    D : float
        Pile diameter [m]
    ra : float, optional
        empirical factor dependent on the number of load cycles, 1.0 for monotonic loading and
        0.3 for cyclic loading at 100 cycles, by default 0.3

    Returns
    -------
    object
        callable for use as p_multipler in `openpile.soilmodels.API_sand` in cyclic mode.

    Example
    -------

    .. plot::

        import matplotlib.pyplot as plt
        from openpile.construct import Layer
        from openpile.utils import hooks
        from openpile.soilmodels import API_sand
        # settings for py curve generation
        kw = {'sig':50, 'X':5, 'layer_height':10, 'depth_from_top_of_layer':5, 'D':5, 'L':30}
        # a, b soil models are traditional cyclic and static API sand models
        a = API_sand(
            phi=33,
            kind="static",
        )
        b = API_sand(
            phi=33,
            kind="cyclic",
        )
        plt.plot(*a.py_spring_fct(**kw), label='static')
        plt.plot(*b.py_spring_fct(**kw), label='cyclic')
        # soil model c is the API sand model with `durkhop` multipliers
        for ra in [0.3,0.45,0.6,0.75,0.9]:
            c = API_sand(
                phi=33,
                kind="cyclic",
                p_multiplier=hooks.durkhop(D=7.0, ra=ra)
            )
            plt.plot(*c.py_spring_fct(**kw), ':',label=f'Durkhop multipliers, ra={ra}')
        plt.legend()

    Reference
    ---------

    .. [Duhr09] Dührkop, J. (2009). *Zum Einfluss von Aufweitungen und zyklischen Lasten auf
        das Verformungsverhalten lateral 385 beanspruchter Pfähle in Sand*. Ph.D. Thesis,
        Institut für Geotechnik und Baubetrieb, Technische Universität Hamburg-Harburg, Vol. 20 (in German).

    """

    func = lambda x: 1 / 0.9 * max(0.9, ra * (3 - 1.143 * x / D) + 0.343 * x / D)

    return func


def durkhop_normalized(D: float, ra: float = 0.3) -> object:
    """This function generates multipliers that represent ratios the cyclic and monotonic curves of
    the traditional API sand model.


    Parameters
    ----------
    D : float
        Pile diameter [m]
    ra : float, optional
        empirical factor dependent on the number of load cycles, 1.0 for monotonic loading and
        0.3 for cyclic loading at 100 cycles, by default 0.3

    Returns
    -------
    object
        callable for use as p_multipler in `openpile.soilmodels.LateralModel` or `openpile.soilmodels.AxialModel`
        The function input is the depth and the function output is the multiplier applied
        for the spring at the said depth.

    Example
    -------


    See also
    --------
    `openpile.utils.hooks.durkhop`
    """

    func = lambda x: 0.9 / max(0.9, ra * (3 - 1.143 * x / D) + 0.343 * x / D)

    return func


class PISA_depth_variation:
    r"""stores functions that returns the depth varying normalized parameters
    of the p-y, m-t, H-b and M-b conic formulations as per [BTZA20]_ and [BHBG20]_, and [BABH20]_.

    These functions can be then be used to model the original models, e.g. Cowden Clay,
    or make modifications to the original models.

    Example
    -------

    .. plot::

        import matplotlib.pyplot as plt
        from openpile.utils.py_curves import custom_pisa_sand, dunkirk_sand
        from openpile.utils.hooks import PISA_depth_variation

        # plot original dunkirk sand curve
        params = {'sig':50, 'X':5, 'Dr':75, 'G0':50e3, 'D':6, 'L':20}
        plt.plot(*dunkirk_sand(**params), label='Original Dunkirk Sand model', linestyle='-')

        # load PISA dunkirk sand depth variation functions
        funcs = PISA_depth_variation.dunkirk_sand_py_pisa_norm_param(D=6, L=20, Dr=75)

        # plot same curve with custom pisa sand curve
        params = {'sig':50, 'G0':50e3, 'D':6, 'X_ult':funcs['py_X'](5), 'Y_ult':funcs['py_Y'](5), 'n':funcs['py_n'](5), 'k':funcs['py_k'](5)}
        plt.plot(*custom_pisa_sand(**params), label='Custom Sand equivalent to Dunkirk sand model', linestyle='--')

        # plot an alrtered curve with custom pisa sand curve
        params['Y_ult'] /= 2
        plt.plot(*custom_pisa_sand(**params), label='Custom Sand with altered ultimate strength', linestyle='--')

        plt.title('Dunkirk Sand model vs. Custom PISA sand model')
        plt.legend()


    """

    @staticmethod
    def dunkirk_sand_pisa_norm_param(D: float, L: float, Dr: float):
        """returns the depth variation functions for all normalized parameters
        of the dunkirk sand conic formulations as per [BTZA20]_.

        Parameters
        ----------
        D : float
            pile diameter [m]
        L : float
            pile embedment [m]
        Dr : float
            sand relative density [%]
        """

        py = PISA_depth_variation.dunkirk_sand_py_pisa_norm_param(D=D, L=L, Dr=Dr)
        mt = PISA_depth_variation.dunkirk_sand_mt_pisa_norm_param(L=L, Dr=Dr)
        Hb = PISA_depth_variation.dunkirk_sand_Hb_pisa_norm_param(D=D, L=L, Dr=Dr)
        Mb = PISA_depth_variation.dunkirk_sand_Mb_pisa_norm_param(D=D, L=L, Dr=Dr)

        return {**py, **mt, **Hb, **Mb}

    @staticmethod
    def cowden_clay_pisa_norm_param(D: float, L: float):
        """returns the depth variation functions for all normalized parameters
        of the cowden_clay conic formulations as per [BHBG20]_.

        Parameters
        ----------
        D : float
            pile diameter [m]
        L : float
            pile embedment [m]
        """

        py = PISA_depth_variation.cowden_clay_py_pisa_norm_param(D=D)
        mt = PISA_depth_variation.cowden_clay_mt_pisa_norm_param(D=D)
        Hb = PISA_depth_variation.cowden_clay_Hb_pisa_norm_param(D=D, L=L)
        Mb = PISA_depth_variation.cowden_clay_Mb_pisa_norm_param(D=D, L=L)

        return {**py, **mt, **Hb, **Mb}

    @staticmethod
    def cowden_clay_py_pisa_norm_param(D: float):
        """returns the depth variation functions for the 4 normalized parameters
        of the cowden_clay p-y conic formulation as per [BHBG20]_.

        Parameters
        ----------
        D : float
            pile diameter [m]

        Returns
        -------
        dict
            dictionary of depth variation functions with keys['py_k','py_n','py_X', 'py_Y']
            accepting one input argument, the depth from ground (strictly positive) [m]
        """
        # normalized initial stiffness
        k_p1 = 10.6
        k_p2 = -1.650
        k_func = lambda x: max(0.0001, k_p1 + k_p2 * x / D)

        # normalized curvature
        n_p1 = 0.9390
        n_p2 = -0.03345
        n_func = lambda x: min(0.999, max(0, n_p1 + n_p2 * x / D))

        # normalized displacement at peak resistance
        X_func = lambda x: 241.4

        # normalized peak resistance
        p_u1 = 10.7
        p_u2 = -7.101
        Y_func = lambda x: max(0.0001, p_u1 + p_u2 * m.exp(-0.3085 * x / D))

        return {"py_k": k_func, "py_n": n_func, "py_X": X_func, "py_Y": Y_func}

    @staticmethod
    def cowden_clay_mt_pisa_norm_param(D: float):
        """returns the depth variation functions for the 4 normalized parameters
        of the cowden_clay rotational spring conic formulation as per [BHBG20]_.

        Parameters
        ----------
        D : float
            pile diameter [m]

        Returns
        -------
        dict
            dictionary of depth variation functions with keys['mt_k','mt_n','mt_X', 'mt_Y']
            accepting one input argument, the depth from ground (strictly positive) [m]
        """
        # normalized initial stiffness
        k_m1 = 1.420
        k_m2 = -0.09643
        k_func = lambda x: max(0.0001, k_m1 + k_m2 * x / D)

        # normalized curvature
        n_func = lambda x: 0.0

        # normalized peak resistance
        m_m1 = 0.2899
        m_m2 = -0.04775
        Y_func = lambda x: max(0.0001, m_m1 + m_m2 * x / D)

        # normalized displacement at peak resistance
        X_func = lambda x: max(0.0001, m_m1 + m_m2 * x / D) / max(0.0001, k_m1 + k_m2 * x / D)

        return {"mt_k": k_func, "mt_n": n_func, "mt_X": X_func, "mt_Y": Y_func}

    @staticmethod
    def cowden_clay_Hb_pisa_norm_param(D: float, L: float):
        """returns the depth variation functions for the 4 normalized parameters
        of the cowden_clay base shear spring conic formulation as per [BHBG20]_.

        Parameters
        ----------
        D : float
            pile diameter [m]
        L : float
            pile embedment length [m]

        Returns
        -------
        dict
            dictionary of depth variation functions with keys['Hb_k','Hb_n','Hb_X', 'Hb_Y']
            accepting one input argument, the depth from ground (strictly positive) [m]
        """
        # normalized initial stiffness
        k_h1 = 2.717
        k_h2 = -0.3575
        k_func = lambda x: max(0.0001, k_h1 + k_h2 * L / D)

        # normalized curvature
        n_h1 = 0.8793
        n_h2 = -0.03150
        n_func = lambda x: min(0.999, max(0, n_h1 + n_h2 * L / D))

        # normalized peak resistance
        p_u1 = 0.4038
        p_u2 = 0.04812
        Y_func = lambda x: max(0.0001, p_u1 + p_u2 * L / D)

        # normalized displacement at peak resistance
        X_func = lambda x: 235.7

        return {"Hb_k": k_func, "Hb_n": n_func, "Hb_X": X_func, "Hb_Y": Y_func}

    @staticmethod
    def cowden_clay_Mb_pisa_norm_param(D: float, L: float):
        """returns the depth variation functions for the 4 normalized parameters
        of the cowden_clay base moment spring conic formulation as per [BHBG20]_.

        Parameters
        ----------
        D : float
            pile diameter [m]
        L : float
            pile embedment length [m]

        Returns
        -------
        dict
            dictionary of depth variation functions with keys['Mb_k','Mb_n','Mb_X', 'Mb_Y']
            accepting one input argument, the depth from ground (strictly positive) [m]
        """
        # normalized initial stiffness
        k_m1 = 0.2146
        k_m2 = -0.002132
        k_func = lambda x: max(0.0001, k_m1 + k_m2 * L / D)

        # normalized curvature
        n_m1 = 1.079
        n_m2 = -0.1087
        n_func = lambda x: min(0.999, max(0, n_m1 + n_m2 * L / D))

        # normalized peak resistance
        m_m1 = 0.8192
        m_m2 = -0.08588
        Y_func = lambda x: max(0.0001, m_m1 + m_m2 * L / D)

        # normalized displacement at peak resistance
        X_func = lambda x: 173.1

        return {"Mb_k": k_func, "Mb_n": n_func, "Mb_X": X_func, "Mb_Y": Y_func}

    @staticmethod
    def dunkirk_sand_py_pisa_norm_param(D: float, L: float, Dr: float):
        """returns the depth variation functions for the 4 normalized parameters
        of the dunkirk_sand p-y conic formulation as per [BTZA20]_.

        Parameters
        ----------
        D : float
            pile diameter [m]
        L : float
            pile embedment length [m]
        Dr : float
            sand relative density [%]

        Returns
        -------
        dict
            dictionary of depth variation functions with keys['py_k','py_n','py_X', 'py_Y']
            accepting one input argument, the depth from ground (strictly positive) [m]
        """
        Dr = Dr / 100

        # normalized initial stiffness
        k_p1 = 8.731 - 0.6982 * Dr
        k_p2 = -0.9178
        k_func = lambda x: max(0.001, k_p1 + k_p2 * x / D)

        # normalized curvature
        n_func = lambda x: min(0.999, max(0, 0.917 + 0.06193 * Dr))

        # normalized displacement at peak resistance
        X_func = lambda x: 146.1 - 92.11 * Dr

        # normalized peak resistance
        p_u1 = 0.3667 + 25.89 * Dr
        p_u2 = 0.3375 - 8.9 * Dr
        Y_func = lambda x: max(0.001, p_u1 + p_u2 * x / L)

        return {"py_k": k_func, "py_n": n_func, "py_X": X_func, "py_Y": Y_func}

    @staticmethod
    def dunkirk_sand_mt_pisa_norm_param(L: float, Dr: float):
        """returns the depth variation functions for the 4 normalized parameters
        of the dunkirk_sand rotational spring conic formulation as per [BTZA20]_.

        Parameters
        ----------
        L : float
            pile embedment length [m]
        Dr : float
            sand relative density [%]

        Returns
        -------
        dict
            dictionary of depth variation functions with keys['mt_k','mt_n','mt_X', 'mt_Y']
            accepting one input argument, the depth from ground (strictly positive) [m]
        """
        Dr = Dr / 100

        # normalized initial stiffness
        k_func = lambda x: 17.0

        # normalized curvature
        n_func = lambda x: 0.0

        # normalized peak resistance
        m_u1 = 0.2605
        m_u2 = -0.1989 + 0.2019 * Dr
        Y_func = lambda x: max(0.001, m_u1 + m_u2 * x / L)

        # normalized displacement at peak resistance
        X_func = lambda x: max(0.001, m_u1 + m_u2 * x / L) / 17.0

        return {"mt_k": k_func, "mt_n": n_func, "mt_X": X_func, "mt_Y": Y_func}

    @staticmethod
    def dunkirk_sand_Hb_pisa_norm_param(D: float, L: float, Dr: float):
        """returns the depth variation functions for the 4 normalized parameters
        of the dunkirk_sand base shear spring conic formulation as per [BTZA20]_.

        Parameters
        ----------
        D : float
            pile diameter [m]
        L : float
            pile embedment length [m]
        Dr : float
            sand relative density [%]

        Returns
        -------
        dict
            dictionary of depth variation functions with keys['Hb_k','Hb_n','Hb_X', 'Hb_Y']
            accepting one input argument, the depth from ground (strictly positive) [m]
        """
        Dr = Dr / 100

        # normalized initial stiffness
        k_h1 = 6.505 - 2.985 * Dr
        k_h2 = -0.007969 - 0.4299 * Dr
        k_func = lambda x: max(0.001, k_h1 + k_h2 * L / D)

        # normalized curvature
        n_h1 = 0.09978 + 0.7974 * Dr
        n_h2 = 0.004994 - 0.07005 * Dr
        n_func = lambda x: min(0.999, max(0.0, n_h1 + n_h2 * L / D))

        # normalized peak resistance
        p_u1 = 0.09952 + 0.7996 * Dr
        p_u2 = 0.03988 - 0.1606 * Dr
        Y_func = lambda x: max(0.001, p_u1 + p_u2 * L / D)

        # normalized displacement at peak resistance
        v_hu1 = 0.5150 + 2.883 * Dr
        v_hu2 = 0.1695 - 0.7018 * Dr
        X_func = lambda x: v_hu1 + v_hu2 * L / D

        return {"Hb_k": k_func, "Hb_n": n_func, "Hb_X": X_func, "Hb_Y": Y_func}

    @staticmethod
    def dunkirk_sand_Mb_pisa_norm_param(D: float, L: float, Dr: float):
        """returns the depth variation functions for the 4 normalized parameters
        of the dunkirk_sand base moment spring conic formulation as per [BTZA20]_.

        Parameters
        ----------
        D : float
            pile diameter [m]
        L : float
            pile embedment length [m]
        Dr : float
            sand relative density [%]

        Returns
        -------
        dict
            dictionary of depth variation functions with keys['Mb_k','Mb_n','Mb_X', 'Mb_Y']
            accepting one input argument, the depth from ground (strictly positive) [m]
        """
        Dr = Dr / 100

        # normalized initial stiffness
        k_func = lambda x: 0.3515

        # normalized curvature
        n_func = lambda x: min(0.999, max(0.0, 0.3 + 0.4986 * Dr))

        # normalized peak resistance
        m_u1 = 0.09981 + 0.3710 * Dr
        m_u2 = 0.01998 - 0.09041 * Dr
        Y_func = lambda x: max(0.001, m_u1 + m_u2 * L / D)

        # normalized displacement at peak resistance
        X_func = lambda x: 44.89

        return {"Mb_k": k_func, "Mb_n": n_func, "Mb_X": X_func, "Mb_Y": Y_func}


class InitialSubgradeReaction:
    """This class stores functions that calculate the initial
    subgrade modulus of soil reaction curves.

    This class includes functions to calculate the initial subgrade reaction of py curves for sand.

    Example
    -------

    .. plot::

        # imports
        from matplotlib import pyplot as plt
        from openpile.utils.py_curves import api_sand
        from openpile.utils.hooks import InitialSubgradeReaction

        # sand parameters
        settings = {'phi':35,'X':5,'sig':50,'D':4, 'below_water_table':False}

        # create a plot
        _, ax = plt.subplots()
        ax.set_title("Varying the initial subgrade reaction modulus")
        # plot API sand
        ax.plot(*api_sand(**settings,), label='API sand')
        # plot API sand with Kallehave modification
        ax.plot(*api_sand(**settings, k=InitialSubgradeReaction.kallehave_sand(**settings)), label="Kallehave sand")
        ax.plot(*api_sand(**settings, k=InitialSubgradeReaction.sørensen2010_sand(**settings)), label="Sørensen sand")
        ax.legend()
    """

    @staticmethod
    def api_sand(phi: float, below_water_table: bool, *args, **kwargs):
        """Calculates the initial subgrade modulus 'k' in the API sand p-y curve.
        The value calculated here is based on a visual fit.

        Parameters
        ----------
        phi : float
            internal angle of friction in degrees
        below_water_table : bool
            whether the curve is below or above the water table

        Returns
        -------
        float
            initial subgrade modulus [kN/m^3]
        """
        if below_water_table:
            return max((0.1978 * phi**2 - 10.232 * phi + 136.82) * 1000, 5400)
        else:
            return max((0.2153 * phi**2 - 8.232 * phi + 63.657) * 1000, 5400)

    @staticmethod
    def kallehave_sand(phi: float, below_water_table: bool, X: float, D: float, *args, **kwargs):
        """Calculates the initial subgrade modulus based on modification of the API sand p-y curve, presented in #REF.

        Parameters
        ----------
        phi : float
            internal angle of friction in degrees
        below_water_table : bool
            whether the curve is below or above the water table
        X : float
            depth from the ground surface [m]
        D : float
            pile diameter [m]

        Returns
        -------
        float
            initial subgrade modulus [kN/m^3]
        """
        return (
            InitialSubgradeReaction.api_sand(phi, below_water_table)
            * (X / 2.5) ** 0.6
            * (D / 0.61) ** 0.5
        )

    @staticmethod
    def sørensen2010_sand(phi: float, X: float, D: float, *args, **kwargs):
        """Calculates the initial subgrade modulus based on modification of the API sand p-y curve,
        presented in [SøIA10]_.

        Parameters
        ----------
        phi : float
            internal angle of friction in degrees
        below_water_table : bool
            whether the curve is below or above the water table
        X : float
            depth from the ground surface [m]
        D : float
            pile diameter [m]

        Returns
        -------
        float
            initial subgrade modulus [kN/m^3]
        """
        Dref = 1.0
        Xref = 1.0
        X = max(X, 0.01)
        return 1 / X * 50e3 * (X / Xref) ** 0.6 * (D / Dref) ** 0.5 * m.radians(phi) ** 3.6



"""
`mt_curves` module
------------------

"""

# Import libraries
import math as m
import numpy as np
from numba import njit, prange
from random import random

from openpile.core.misc import conic

# SPRING FUNCTIONS --------------------------------------------
@njit(cache=True)
def bothkennar_clay(
    X: float,
    Su: float,
    G0: float,
    D: float,
    output_length: int = 20,
):
    """
    Create the rotational springs from the PISA clay formulation
    published by Burd et al 2020 (see [BABH20]_) and calibrated based on Bothkennar clay
    response (a normally consolidated soft clay).

    Parameters
    ----------
    X : float
        Depth below ground level [unit: m]
    Su : float
        Undrained shear strength [unit: kPa]
    G0 : float
        Small-strain shear modulus [unit: kPa]
    D : float
        Pile diameter [unit: m]
    output_length : int, optional
        Number of datapoints in the curve, by default 20

    Returns
    -------
    1darray
        t vector of length [unit: rad]
    1darray
        m vector [unit: kN]

    """

    # Bothkennar clay parameters
    k_m1 = 1.698
    k_m2 = -0.1576
    n_m1 = 0.00
    n_m2 = 0.00
    m_m1 = 0.4862
    m_m2 = -0.05674

    # Depth variation parameters
    k = k_m1 + k_m2 * X / D
    n = n_m1 + n_m2 * X / D
    m_max = m_m1 + m_m2 * X / D
    psi_max = m_max / k

    # calculate normsalised conic function
    t, m = conic(psi_max, n, k, m_max, output_length)

    # return non-normalised curve
    return t * (Su / G0), m * (Su * D**2)


@njit(cache=True)
def cowden_clay(
    X: float,
    Su: float,
    G0: float,
    D: float,
    output_length: int = 20,
):
    """
    Create the rotational springs from the PISA clay formulation
    published by Byrne et al 2020 (see [BHBG20]_) and calibrated based pile
    load tests at Cowden (north east coast of England).

    Parameters
    ----------
    X : float
        Depth below ground level [unit: m]
    Su : float
        Undrained shear strength [unit: kPa]
    G0 : float
        Small-strain shear modulus [unit: kPa]
    D : float
        Pile diameter [unit: m]
    output_length : int, optional
        Number of datapoints in the curve, by default 20

    Returns
    -------
    1darray
        t vector of length [unit: rad]
    1darray
        m vector [unit: kN]

    """

    # Cowden clay parameters
    k_m1 = 1.420
    k_m2 = -0.09643
    n_m1 = 0.00
    n_m2 = 0.00
    m_m1 = 0.2899
    m_m2 = -0.04775

    # Depth variation parameters
    k = k_m1 + k_m2 * X / D
    n = n_m1 + n_m2 * X / D
    m_max = m_m1 + m_m2 * X / D
    psi_max = m_max / k

    # calculate normsalised conic function
    t, m = conic(psi_max, n, k, m_max, output_length)

    # return non-normalised curve
    return t * (Su / G0), m * (Su * D**2)


@njit(cache=True)
def dunkirk_sand(
    sig: float,
    X: float,
    Dr: float,
    G0: float,
    p: float,
    D: float,
    L: float,
    output_length: int = 20,
):
    """
    Create the rotational springs from the PISA sand
    formulation published by Burd et al (2020) (see [BTZA20]_).
    Also called the General Dunkirk Sand Model (GDSM)

    Parameters
    ----------
    sig : float
        vertical/overburden effective stress [unit: kPa]
    X : float
        Depth below ground level [unit: m]
    Dr : float
        Sand relative density Value must be between 0 and 100 [unit: -]
    G0 : float
        Small-strain shear modulus [unit: kPa]
    p : float
        radial stress computed via p-y curves [unit: kN/m]
    D : float
        Pile diameter [unit: m]
    L : float
        Embedded pile length [unit: m]
    output_length : int, optional
        Number of datapoints in the curve, by default 20

    Returns
    -------
    1darray
        t vector of length [unit: rad]
    1darray
        m vector [unit: kN]
    """
    # correct relative density for decimal value
    Dr = Dr / 100

    # correct p value
    p = abs(p)

    # Generalised Dunkirk Sand Model parameters
    k_m1 = 17.00
    k_m2 = 0.00
    n_m = 0.0
    m_u1 = 0.2605
    m_u2 = -0.1989 + 0.2019 * Dr

    # Depth variation parameters
    k = k_m1 + k_m2 * X / D
    n = n_m
    m_max = m_u1 + m_u2 * X / L
    psi_max = m_max / k

    # calculate normsalised conic function
    t, m = conic(psi_max, n, k, m_max, output_length)

    # return non-normalised curve
    return t * (sig / G0), m * (p * D)


@njit(cache=True)
def custom_pisa_sand(
    sig: float,
    G0: float,
    p: float,
    D: float,
    X_ult: float,
    n: float,
    k: float,
    Y_ult: float,
    output_length: int = 20,
):
    """Creates a rotational spring with the PISA sand formulation and custom user inputs.

    Parameters
    ----------
    sig : float
        vertical/overburden effective stress [unit: kPa]
    G0 : float
        Small-strain shear modulus [unit: kPa]
    p : float
        Radial stress computed via p-y curves [unit: kN/m]
    D : float
        Pile diameter [m]
    X_ult : float
        Normalized displacement at maximum strength
    k : float
        Normalized stiffness parameter
    n : float
        Normalized curvature parameter, must be between 0 and 1
    Y_ult : float
        Normalized maximum strength parameter
    output_length : int, optional
        Number of datapoints in the curve, by default 20

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]
    """

    # calculate normsalised conic function
    x, y = conic(X_ult, n, k, Y_ult, output_length)

    # return non-normalised curve
    return x * (sig / G0), y * (p * D)


@njit(cache=True)
def custom_pisa_clay(
    Su: float,
    G0: float,
    D: float,
    X_ult: float,
    n: float,
    k: float,
    Y_ult: float,
    output_length: int = 20,
):
    """
    Creates a rotational spring with the PISA clay formulation and custom user inputs.


    Parameters
    ----------
    Su : float
        Undrained shear strength [unit: kPa]
    G0 : float
        Small-strain shear modulus [unit: kPa]
    D : float
        Pile diameter [unit: m]
    X_ult : float
        Normalized displacement at maximum strength
    k : float
        Normalized stiffness parameter
    n : float
        Normalized curvature parameter, must be between 0 and 1
    Y_ult : float
        Normalized maximum strength parameter
    output_length : int, optional
        Number of datapoints in the curve, by default 20

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]

    """
    # calculate normsalised conic function
    x, y = conic(X_ult, n, k, Y_ult, output_length)

    # return non-normalised curve
    return x * (Su / G0), y * (Su * D**2)



from openpile import construct
from openpile.soilmodels import API_clay, API_sand, API_clay_axial, API_sand_axial
from openpile.materials import PileMaterial

import pytest
import numpy as np
import math as m

from pydantic import ValidationError


@pytest.fixture
def offshore_wind_pile1():
    # create a steel and circular pile
    return construct.Pile(
        name="",
        material="Steel",
        sections=[
            construct.CircularPileSection(top=0, bottom=-10, diameter=7.5, thickness=0.07),
            construct.CircularPileSection(top=-10, bottom=-40, diameter=7.5, thickness=0.08),
        ],
    )


@pytest.fixture
def circular_slender_pile():
    # create a steel and circular pile
    return construct.Pile(
        name="",
        material="Steel",
        sections=[
            construct.CircularPileSection(top=0, bottom=-10, diameter=0.7, thickness=0.02),
            construct.CircularPileSection(top=-10, bottom=-40, diameter=0.7, thickness=0.02),
        ],
    )


@pytest.fixture
def circular_slender_pile():
    # create a steel and circular pile
    return construct.Pile(
        name="",
        material="Steel",
        sections=[
            construct.CircularPileSection(top=0, bottom=-10, diameter=0.7, thickness=0.02),
            construct.CircularPileSection(top=-10, bottom=-40, diameter=0.7, thickness=0.02),
        ],
    )


class TestPileSection:
    def test_pile_section(self):
        section = construct.CircularPileSection(top=10, bottom=5, diameter=9.4, thickness=0.31)
        assert section.width == 9.4
        assert section.top == 10
        assert section.bottom == 5
        assert section.length == 5

    def test_elevations(self):
        with pytest.raises(ValueError):
            construct.CircularPileSection(top=-10, bottom=5, diameter=9.4, thickness=0.31)

    def test_default_thickness(self):
        section = construct.CircularPileSection(top=10, bottom=5, diameter=9)
        assert section.thickness == 4.5


class TestPile:
    def test_main_constructor(self, offshore_wind_pile1):
        """Main constructor of Pile object, check if pile.data
        has even entries and that steel E value is 210GPa
        """

        # check Young modulus is indeed Steel
        assert offshore_wind_pile1.E == 210e6
        # check even numbers of row for dataframe
        assert offshore_wind_pile1.data.values.shape[0] % 2 == 0

    def test_custom_material(self):
        pile = construct.Pile(
            name="",
            material="Steel",
            sections=[
                construct.CircularPileSection(top=0.1, bottom=-9.9, diameter=8.0, thickness=0.07),
            ],
        )

        assert pile.material.name == "Steel"
        assert isinstance(pile.material, PileMaterial)

    def test_pile_width(self):
        pile = construct.Pile(
            name="",
            material="Steel",
            sections=[
                construct.CircularPileSection(top=0.1, bottom=-9.9, diameter=8.0, thickness=0.07),
                construct.CircularPileSection(top=-9.9, bottom=-39.9, diameter=8.0, thickness=0.08),
            ],
        )

        assert pile.sections[0].width == 8.0
        assert pile.sections[1].width == 8.0

    def test_pile_length(self):
        pile = construct.Pile(
            name="",
            material="Steel",
            sections=[
                construct.CircularPileSection(top=10, bottom=-12, diameter=7.5, thickness=0.07),
                construct.CircularPileSection(top=-12, bottom=-28, diameter=8.5, thickness=0.08),
            ],
        )

        assert pile.length == 38.0

    def test_pile_bottom(self):
        pile = construct.Pile(
            name="",
            material="Steel",
            sections=[
                construct.CircularPileSection(top=10, bottom=-12, diameter=7.5, thickness=0.07),
                construct.CircularPileSection(top=-12, bottom=-28, diameter=8.5, thickness=0.08),
            ],
        )

        assert pile.bottom_elevation == -28.0

    def test_pile_area(self):
        """check pile area"""
        # Create a pile instance with two sections of respectively 10m and 30m length.
        pile = construct.Pile(
            name="",
            material="Steel",
            sections=[
                construct.CircularPileSection(top=10, bottom=0, diameter=1.0, thickness=0.5),
                construct.CircularPileSection(top=0, bottom=-30, diameter=1.0, thickness=0.5),
            ],
        )

        assert pile.sections[0].area == 0.25 * m.pi
        assert pile.sections[1].area == 0.25 * m.pi
        assert pile.tip_area == 0.25 * m.pi
        assert pile.tip_footprint == 0.25 * m.pi


class TestLayer:
    def test_constructor(self):
        try:
            layer = construct.Layer(
                name="Soft Clay",
                top=0,
                bottom=-10,
                weight=19,
                lateral_model=API_clay(Su=[30, 35], eps50=[0.01, 0.02], kind="cyclic"),
            )
        except Exception:
            assert False, "Constructor did not work"

    def test_wrong_order_top_and_bottom(self):
        with pytest.raises(ValidationError):
            layer = construct.Layer(
                name="Soft Clay",
                top=-10,
                bottom=0,
                weight=19,
                lateral_model=API_clay(Su=[30, 35], eps50=[0.01, 0.02], kind="cyclic"),
            )

    def test_equal_top_and_bottom(self):
        with pytest.raises(ValidationError):
            layer = construct.Layer(
                name="Soft Clay",
                top=-5,
                bottom=-5,
                weight=19,
                lateral_model=API_clay(Su=[30, 35], eps50=[0.01, 0.02], kind="cyclic"),
            )

    def test_wrong_unit_weight(self):
        """Constructor should not accept a soil unit weight equal to 10 or less as
        it would be less dense than water and likely the user made a mistake with
        the effective unit weight.
        """
        with pytest.raises(ValidationError):
            layer = construct.Layer(
                name="Soft Clay",
                top=0,
                bottom=-10,
                weight=10,
                lateral_model=API_clay(Su=[30, 35], eps50=[0.01, 0.02], kind="static"),
            )


class TestSoilProfile:
    def test_wrong_layers_elevations_gap(self):
        with pytest.raises(ValidationError):
            sp = construct.SoilProfile(
                name="",
                top_elevation=0,
                water_line=0,
                layers=[
                    construct.Layer(
                        name="",
                        top=0,
                        bottom=-10,
                        weight=20,
                    ),
                    construct.Layer(
                        name="",
                        top=-15,
                        bottom=-30,
                        weight=20,
                    ),
                ],
            )

    def test_wrong_layers_elevations_overlap(self):
        with pytest.raises(ValidationError):
            sp = construct.SoilProfile(
                name="",
                top_elevation=0,
                water_line=0,
                layers=[
                    construct.Layer(
                        name="",
                        top=0,
                        bottom=-10,
                        weight=20,
                    ),
                    construct.Layer(
                        name="",
                        top=-8,
                        bottom=-30,
                        weight=20,
                    ),
                ],
            )

    def test_wrong_top_elevation(self):
        with pytest.raises(ValidationError):
            sp = construct.SoilProfile(
                name="",
                top_elevation=10,
                water_line=0,
                layers=[
                    construct.Layer(
                        name="",
                        top=11,
                        bottom=-10,
                        weight=20,
                    ),
                    construct.Layer(
                        name="",
                        top=-10,
                        bottom=-30,
                        weight=20,
                    ),
                ],
            )

    def test_wrong_bottom_elevation(self):
        sp = construct.SoilProfile(
            name="",
            top_elevation=10,
            water_line=0,
            layers=[
                construct.Layer(
                    name="",
                    top=10,
                    bottom=-10,
                    weight=20,
                ),
                construct.Layer(
                    name="",
                    top=-10,
                    bottom=-30,
                    weight=20,
                ),
            ],
        )

        assert sp.bottom_elevation == -30


class TestModel:
    def test_boundary_conditions_simple_beam(self, circular_slender_pile):
        """this test should show that specific nodes are created where x2mesh is used."""
        model = construct.Model(
            name="",
            pile=circular_slender_pile,
            boundary_conditions=[
                construct.BoundaryFixation(
                    elevation=circular_slender_pile.top_elevation, y=True, z=True
                ),
                construct.BoundaryFixation(
                    elevation=circular_slender_pile.bottom_elevation, y=True
                ),
                construct.BoundaryDisplacement(
                    elevation=0.5
                    * (
                        circular_slender_pile.top_elevation + circular_slender_pile.bottom_elevation
                    ),
                    y=0.1,
                ),
            ],
        )

        results = model.solve()
        assert results.deflection["Deflection [m]"].max() == 0.1

    def test_boundary_conditions_fixed_end_beam(self, circular_slender_pile):
        """this test checks that the beam calculation with one end fixed and
        the other end loaded by a transverse force results in the correct known solution"""
        model = construct.Model(
            name="",
            pile=circular_slender_pile,
            boundary_conditions=[
                construct.BoundaryFixation(
                    elevation=circular_slender_pile.bottom_elevation, x=True, y=True, z=True
                ),
                construct.BoundaryForce(elevation=circular_slender_pile.top_elevation, y=1.0),
            ],
        )

        results = model.solve()
        assert m.isclose(
            results.deflection["Deflection [m]"][0],
            circular_slender_pile.length**3
            / (
                3
                * circular_slender_pile.E
                * circular_slender_pile.sections[0].second_moment_of_area
            ),
            abs_tol=1e-5,
        )

    def test_constructor_with_x2mesh(self, circular_slender_pile):
        """this test should show that specific nodes are created where x2mesh is used."""
        model = construct.Model(name="", pile=circular_slender_pile, x2mesh=[-10.38, -10.918])

        assert any(model.nodes_coordinates["z [m]"] == -10.38)
        assert any(model.nodes_coordinates["z [m]"] == -10.918)

    def test_constructor_wo_soil_models(self):
        """
        check that model can still be created with no lateral or axial models"""
        try:
            model = construct.Model(
                name="",
                pile=construct.Pile.create_tubular(
                    name="",
                    top_elevation=20,
                    bottom_elevation=-30,
                    material="Steel",
                    diameter=7,
                    wt=0.08,
                ),
                soil=construct.SoilProfile(
                    name="",
                    top_elevation=10,
                    water_line=0,
                    layers=[
                        construct.Layer(
                            name="",
                            top=10,
                            bottom=-10,
                            weight=20,
                        ),
                        construct.Layer(
                            name="",
                            top=-10,
                            bottom=-30,
                            weight=20,
                        ),
                    ],
                ),
            )
        except Exception:
            assert False, f"Constructor does not work"

        # check that p_y springs are all zero
        assert not np.all(
            model.get_distributed_lateral_springs()[["VAL 0", "VAL 1", "VAL 2"]].values
        )

        # # check that m_t springs are all zero
        assert not np.all(
            model.get_distributed_rotational_springs()[["VAL 0", "VAL 1", "VAL 2"]].values
        )

        assert not np.all(model.get_base_rotational_spring()[["VAL 0", "VAL 1", "VAL 2"]].values)

        assert not np.all(model.get_base_shear_spring()[["VAL 0", "VAL 1", "VAL 2"]].values)

    def test_entrapped_soil_weight_above_water_table(self):
        """calculate the weight of the soil inside a pile that is above water table"""

        # the special diameter and wall thickness is calculated and applied such that
        # a metre long of pile with this diameter ie quivalent
        # to one cubic metre
        special_diameter = (4 / m.pi) ** 0.5
        special_wallthickness = 0.01
        soil_weight = 18

        # a pile with the special diameter and an unreasonably thin wall thickness
        p = construct.Pile.create_tubular(
            name="<pile name>",
            top_elevation=0,
            bottom_elevation=-10,
            diameter=special_diameter + (2 * special_wallthickness),
            wt=special_wallthickness,
        )

        # Create a 40m deep offshore Soil Profile with a 15m water column
        sp = construct.SoilProfile(
            name="Offshore Soil Profile",
            top_elevation=0,
            water_line=-15,
            layers=[
                construct.Layer(
                    name="medium dense sand",
                    top=0,
                    bottom=-20,
                    weight=soil_weight,
                    lateral_model=API_sand(
                        phi=33,
                        kind="cyclic",
                        extension="mt_curves",
                    ),
                ),
            ],
        )

        model = construct.Model(name="", pile=p, soil=sp)
        # check
        assert m.isclose(model.entrapped_soil_weight, soil_weight * p.length)

    def test_entrapped_soil_weight_below_water_table(self):
        """calculate the weight of the soil inside a pile that is submerged in water"""

        # the special diameter and wall thickness is calculated and applied such that
        # a metre long of pile with this diameter ie quivalent
        # to one cubic metre
        special_diameter = (4 / m.pi) ** 0.5
        special_wallthickness = 0.01
        soil_weight = 18

        # a pile with the special diameter and an unreasonably thin wall thickness
        p = construct.Pile.create_tubular(
            name="<pile name>",
            top_elevation=0,
            bottom_elevation=-10,
            diameter=special_diameter + (2 * special_wallthickness),
            wt=special_wallthickness,
        )

        # Create a 40m deep offshore Soil Profile with a 15m water column
        sp = construct.SoilProfile(
            name="Offshore Soil Profile",
            top_elevation=0,
            water_line=10,
            layers=[
                construct.Layer(
                    name="medium dense sand",
                    top=0,
                    bottom=-20,
                    weight=soil_weight,
                    lateral_model=API_sand(
                        phi=33,
                        kind="cyclic",
                        extension="mt_curves",
                    ),
                ),
            ],
        )

        model = construct.Model(name="", pile=p, soil=sp)
        # check
        assert m.isclose(model.entrapped_soil_weight, (soil_weight - 10) * p.length)

    def test_submerged_effective_pile_weight(self):

        #  the special diameter and wall thickness is calculated and applied such that
        # a metre long of pile with this diameter ie quivalent
        # to one cubic metre
        special_diameter = 10 / m.pi
        special_wallthickness = 0.001
        steel_weight = 78

        # a pile with the special diameter and an unreasonably thin wall thickness
        p = construct.Pile.create_tubular(
            name="<pile name>",
            top_elevation=0,
            bottom_elevation=-100,
            diameter=special_diameter,
            wt=special_wallthickness,
        )

        # Create a 40m deep offshore Soil Profile with a 15m water column
        sp = construct.SoilProfile(
            name="Offshore Soil Profile",
            top_elevation=0,
            water_line=15,
            layers=[
                construct.Layer(
                    name="medium dense sand",
                    top=0,
                    bottom=-100,
                    weight=18,
                    lateral_model=API_sand(
                        phi=33,
                        kind="cyclic",
                        extension="mt_curves",
                    ),
                ),
            ],
        )

        model = construct.Model(name="", pile=p, soil=sp)

        # check
        assert m.isclose(model.effective_pile_weight, (steel_weight - 10), abs_tol=0.1)

    def test_half_submerged_effective_pile_weight(self):

        #  the special diameter and wall thickness is calculated and applied such that
        # a metre long of pile with this diameter ie quivalent
        # to one cubic metre
        special_diameter = 10 / m.pi
        special_wallthickness = 0.001
        steel_weight = 78

        # a pile with the special diameter and an unreasonably thin wall thickness
        p = construct.Pile.create_tubular(
            name="<pile name>",
            top_elevation=0,
            bottom_elevation=-100,
            diameter=special_diameter,
            wt=special_wallthickness,
        )

        # Create a 40m deep offshore Soil Profile with a 15m water column
        sp = construct.SoilProfile(
            name="Offshore Soil Profile",
            top_elevation=0,
            water_line=-50,
            layers=[
                construct.Layer(
                    name="medium dense sand",
                    top=0,
                    bottom=-100,
                    weight=18,
                    lateral_model=API_sand(
                        phi=33,
                        kind="cyclic",
                        extension="mt_curves",
                    ),
                ),
            ],
        )

        model = construct.Model(name="", pile=p, soil=sp)

        # check
        target_weight = 0.5 * ((steel_weight - 10) + steel_weight)
        assert m.isclose(model.effective_pile_weight, target_weight, abs_tol=0.1)

    def test_axial_capacity_same_as_winkler(self):

        p = construct.Pile.create_tubular(
            name="", top_elevation=0, bottom_elevation=-20, diameter=7.5, wt=0.075
        )
        # Create a 40m deep offshore Soil Profile with a 15m water column
        sp = construct.SoilProfile(
            name="Offshore Soil Profile",
            top_elevation=0,
            water_line=15,
            layers=[
                construct.Layer(
                    name="medium dense sand",
                    top=0,
                    bottom=-20,
                    weight=18,
                    axial_model=API_sand_axial(delta=28),
                ),
            ],
        )
        # Create Model
        M = construct.Model(name="Settlement", pile=p, soil=sp)
        # Apply bottom fixity along lateral axis
        M.set_support(elevation=-20, Ty=True)
        M.set_support(elevation=0, Ty=True)
        # Apply axial and lateral loads
        M.set_pointdisplacement(elevation=0, Tz=-1)
        # Run analysis
        result = M.solve()

        # test
        assert m.isclose(
            -(M.tip_resistance + M.shaft_resistance[0]), result.forces["N [kN]"][0], rel_tol=1e-3
        )



from setuptools import setup
from pathlib import Path
import re

if __name__ == "__main__":
    version = re.findall(
        r'__version__ = "(\d+.\d+.\d+[^"]*)"',
        Path("src/openpile/__init__.py").read_text(encoding="utf-8"),
    )[0]

    setup(
        version=version,
        long_description=Path("README.md").read_text(encoding="utf-8"),
        long_description_content_type="text/markdown",
        project_urls={
            "Documentation": "https://openpile.readthedocs.io/en/latest/",
            "Code": "https://github.com/TchilDill/openpile",
            "Issue tracker": "https://github.com/TchilDill/openpile/issues",
        },
    )



r"""
`winkler` module
==================

The `winkler` module is used to run 1D Finite Element analyses. 

"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import warnings

from dataclasses import dataclass
from copy import deepcopy

from openpile.core import kernel
import openpile.core.validation as validation
import openpile.utils.graphics as graphics
import openpile.core.misc as misc


class PydanticConfig:
    arbitrary_types_allowed = True
    frozen = True
    underscore_attrs_are_private = True


def springs_mob_to_df(model, d):

    # elevations
    x = model.nodes_coordinates["z [m]"].values
    x = kernel.double_inner_njit(x)

    # PY springs
    # py secant stiffness
    py_ks = kernel.calculate_py_springs_stiffness(
        u=d[1::3], springs=model._py_springs, kind="secant"
    ).flatten()
    py_mob = kernel.double_inner_njit(d[1::3]) * py_ks
    # calculate max spring values
    py_max = model._py_springs[:, :, 0].max(axis=2).flatten()

    # mt springs
    # mt secant stiffness
    mt_ks = kernel.calculate_mt_springs_stiffness(
        d[2::3], model._mt_springs, model._py_springs, py_mob.reshape((-1, 2, 1, 1)), kind="secant"
    ).flatten()
    mt_mob = kernel.double_inner_njit(d[2::3]) * mt_ks
    # calculate max spring values
    mt_max = model._mt_springs[:, :, 0, -1].max(axis=2).flatten()

    # create DataFrame
    df = pd.DataFrame(
        data={
            "Elevation [m]": x,
            "p_mobilized [kN/m]": np.abs(py_mob),
            "p_max [kN/m]": py_max,
            "m_mobilized [kNm/m]": np.abs(mt_mob),
            "m_max [kNm/m]": mt_max,
        }
    )

    return df


def reaction_forces_to_df(model, Q):
    z = model.nodes_coordinates["z [m]"].values
    Q = Q.reshape(-1, 3)

    df = pd.DataFrame(
        data={
            "Elevation [m]": z,
            "Nr [kN]": Q[:, 0],
            "Vr [kN]": Q[:, 1],
            "Mr [kNm]": Q[:, 2],
        }
    )
    df[["Nr [kN]"]] = df[["Nr [kN]"]].mask(df[["Nr [kN]"]].abs() < 1e-3, 0.0)
    df[["Vr [kN]"]] = df[["Vr [kN]"]].mask(df[["Vr [kN]"]].abs() < 1e-3, 0.0)
    df[["Mr [kNm]"]] = df[["Mr [kNm]"]].mask(df[["Mr [kNm]"]].abs() < 1e-3, 0.0)

    return df[np.any(df[["Nr [kN]", "Vr [kN]", "Mr [kNm]"]].abs() > 1e-3, axis=1)].reset_index(
        drop=True
    )


def structural_forces_to_df(model, q):
    z = model.nodes_coordinates["z [m]"].values
    z = misc.repeat_inner(z)
    L = kernel.mesh_to_element_length(model).reshape(-1)

    N = np.vstack((-q[0::6], q[3::6])).reshape(-1, order="F")
    V = np.vstack((-q[1::6], q[4::6])).reshape(-1, order="F")
    M = np.vstack((-q[2::6], -q[2::6] + L * q[1::6])).reshape(-1, order="F")

    structural_forces_to_DataFrame = pd.DataFrame(
        data={
            "Elevation [m]": z,
            "N [kN]": N,
            "V [kN]": V,
            "M [kNm]": M,
        }
    )

    return structural_forces_to_DataFrame


def disp_to_df(model, u):
    z = model.nodes_coordinates["z [m]"].values

    Tz = u[::3].reshape(-1)
    Ty = u[1::3].reshape(-1)
    Rx = u[2::3].reshape(-1)

    disp_to_DataFrame = pd.DataFrame(
        data={
            "Elevation [m]": z,
            "Settlement [m]": Tz,
            "Deflection [m]": Ty,
            "Rotation [rad]": Rx,
        }
    )

    return disp_to_DataFrame


@dataclass
class WinklerResult:
    """The `WinklerResult` class is created by any analyses from the :py:mod:`openpile.winkler` module.

    As such the user can use the following properties and/or methods for any return values of an analysis.

    """

    _name: str
    _d: pd.DataFrame
    _f: pd.DataFrame
    _Q: pd.DataFrame
    _dist_mob: pd.DataFrame = None
    _hb_mob: tuple = None
    _mb_mob: tuple = None
    _details: dict = None

    @property
    def displacements(self):
        """Retrieves displacements along each dimensions

        Returns
        -------
        pandas.DataFrame
            Table with the nodes elevations along the pile and their displacements
        """
        return self._d

    @property
    def forces(self):
        """Retrieves forces along the pile (Normal force, shear force and bending moment)

        Returns
        -------
        pandas.DataFrame
            Table with the nodes elevations along the pile and their forces
        """
        return self._f

    @property
    def reactions(self):
        """Retrieves reaction forces (where supports are given)

        Returns
        -------
        pandas.DataFrame
            Table with the nodes elevations along the pile and their forces
        """
        return self._Q

    @property
    def settlement(self):
        """Retrieves degrees of freedom for settlement

        Returns
        -------
        pandas.DataFrame
            Table with the nodes elevations along the pile and their normal displacements
        """
        return self._d[["Elevation [m]", "Settlement [m]"]]

    @property
    def deflection(self):
        """Retrieves degrees of freedom for deflection

        Returns
        -------
        pandas.DataFrame
            Table with the nodes elevations along the pile and their transversal displacements
        """
        return self._d[["Elevation [m]", "Deflection [m]"]]

    @property
    def rotation(self):
        """Retrieves rotational degrees of freedom

        Returns
        -------
        pandas.DataFrame
            Table with the nodes elevations along the pile and their rotations
        """
        return self._d[["Elevation [m]", "Rotation [rad]"]]

    @property
    def py_mobilization(self):
        """Retrieves mobilized resistance of districuted lateral p-y curves.

        Returns
        -------
        pandas.DataFrame
            Table with the nodes elevations along the pile and the mobilized resistance in kN/m.
        """

        if self._dist_mob is None:
            return None
        else:
            return self._dist_mob[["Elevation [m]", "p_mobilized [kN/m]", "p_max [kN/m]"]]

    @property
    def mt_mobilization(self):
        """Retrieves mobilized resistance of distributed moment rotational curves.

        Returns
        -------
        pandas.DataFrame
            Table with the nodes elevations along the pile and the mobilized resistance in kNm/m.
        """
        if self._dist_mob is None:
            return None
        else:
            return self._dist_mob[["Elevation [m]", "m_mobilized [kNm/m]", "m_max [kNm/m]"]]

    @property
    def Hb_mobilization(self):
        """Retrieves mobilized resistance of base shear.

        Returns
        -------
        tuple
            the mobilised value and the maximum resistance in kN
        """
        return self._hb_mob if self._hb_mob is not None else None

    @property
    def Mb_mobilization(self):
        """Retrieves mobilized resistance of base moment.

        Returns
        -------
        tuple
            the mobilised value and the maximum resistance in kNm
        """
        return self._mb_mob if self._mb_mob is not None else None

    def plot_deflection(self, assign=False):
        r"""
        Plots the deflection of the pile.

        Parameters
        ----------
        assign : bool, optional
            by default False

        Returns
        -------
        None or matplotlib.pyplot.figure
            if assign is True, a matplotlib figure is returned

        Example
        -------
        The plot looks like:

        .. plot::
            :context: reset
            :include-source: False

            from openpile.construct import Pile, SoilProfile, Layer, Model
            from openpile.soilmodels import API_clay, API_sand
            p = Pile.create_tubular(
                name="<pile name>", top_elevation=0, bottom_elevation=-40, diameter=7.5, wt=0.075
            )
            # Create a 40m deep offshore Soil Profile with a 15m water column
            sp = SoilProfile(
                name="Offshore Soil Profile",
                top_elevation=0,
                water_line=15,
                layers=[
                    Layer(
                        name="medium dense sand",
                        top=0,
                        bottom=-20,
                        weight=18,
                        lateral_model=API_sand(phi=33, kind="cyclic"),
                    ),
                    Layer(
                        name="firm clay",
                        top=-20,
                        bottom=-40,
                        weight=18,
                        lateral_model=API_clay(Su=[50, 70], eps50=0.015, kind="cyclic"),
                    ),
                ],
            )
            # Create Model
            M = Model(name="<model name>", pile=p, soil=sp)
            # Apply bottom fixity along z-axis
            M.set_support(elevation=-40, Tz=True)
            # Apply axial and lateral loads
            M.set_pointload(elevation=0, Pz=-20e3, Py=5e3)
            # Run analysis
            result = M.solve()
            # plot the results
            result.plot_deflection()

        """
        fig = graphics.plot_deflection(self)
        return fig if assign else None

    def plot_forces(self, assign=False):
        r"""Plots the pile sectional forces.

        Parameters
        ----------
        assign : bool, optional
            by default False

        Returns
        -------
        None or matplotlib.pyplot.figure
            if assign is True, a matplotlib figure is returned

        Example
        -------
        The plot looks like:

        .. plot::
            :context: close-figs
            :include-source: False

            result.plot_forces()
        """
        fig = graphics.plot_forces(self)
        return fig if assign else None

    def plot_lateral_results(self, assign=False):
        r"""Plots the pile deflection and sectional forces.

        Parameters
        ----------
        assign : bool, optional
            by default False

        Returns
        -------
        None or matplotlib.pyplot.figure
            if assign is True, a matplotlib figure is returned


        Example
        -------
        The plot looks like:

        .. plot::
            :context: close-figs
            :include-source: False

            result.plot_lateral_results()
        """
        fig = graphics.plot_results(self)
        return fig if assign else None

    def plot_axial_results(self, assign=False):
        r"""Plots the pile settlements and normal forces.

        Parameters
        ----------
        assign : bool, optional
            by default False

        Returns
        -------
        None or matplotlib.pyplot.figure
            if assign is True, a matplotlib figure is returned

        """
        fig = graphics.plot_settlement(self)
        return fig if assign else None

    def plot(self, assign=False):
        r"""Same behaviour as :py:meth:`openpile.analyze.plot_lateral_results`.

        Parameters
        ----------
        assign : bool, optional
            by default False

        Returns
        -------
        None or matplotlib.pyplot.figure
            if assign is True, a matplotlib figure is returned

        Example
        -------

        The plot looks like:

        .. plot::
            :context: close-figs
            :include-source: False

            result.plot()
        """
        return self.plot_lateral_results(assign)

    def details(self) -> dict:
        """Provide a summary of the results.

        Returns
        -------
        dict
            info on the results

        """

        return {
            **self._details,
            "Max. normal force [kN]": round(self._f["N [kN]"].max(), 2),
            "Min. normal force [kN]": round(self._f["N [kN]"].min(), 2),
            "Max. shear force [kN]": round(self._f["V [kN]"].max(), 2),
            "Min. shear force [kN]": round(self._f["V [kN]"].min(), 2),
            "Max. moment [kNm]": round(self._f["M [kNm]"].max(), 2),
            "Min. moment [kNm]": round(self._f["M [kNm]"].min(), 2),
            "Max. settlement [m]": round(self._d["Settlement [m]"].max(), 3),
            "Min. settlement [m]": round(self._d["Settlement [m]"].min(), 3),
            "Max. deflection [m]": round(self._d["Deflection [m]"].max(), 3),
            "Min. deflection [m]": round(self._d["Deflection [m]"].min(), 3),
            "Max. rotation [rad]": round(self._d["Rotation [rad]"].max(), 3),
            "Min. rotation [rad]": round(self._d["Rotation [rad]"].min(), 3),
        }


def beam(model):
    """
    Function where loading or displacement defined in the model boundary conditions
    are used to solve the system of equations, this is a linear problem and is solved with one iteration.

    Parameters
    ----------
    model : `openpile.construct.Model` object
        Model where structure and boundary conditions are defined.

    Returns
    -------
    results : `openpile.compute.Result` object
        Results of the analysis
    """

    # validate boundary conditions
    # validation.check_boundary_conditions(model)

    # initialise global force
    F = kernel.mesh_to_global_force_dof_vector(model.global_forces)
    # initiliase prescribed displacement vector
    U = kernel.mesh_to_global_disp_dof_vector(model.global_disp)
    # initialise global supports vector
    supports = (kernel.mesh_to_global_restrained_dof_vector(model.global_restrained)) | (U != 0.0)

    # initialise displacement vectors
    d = np.zeros(U.shape)

    # initialise global stiffness matrix
    K = kernel.build_stiffness_matrix(model, d)
    # first run with no stress stiffness matrix
    d, Q = kernel.solve_equations(K, F, U, restraints=supports)

    # internal forces
    q_int = kernel.pile_internal_forces(model, d)

    # Final results
    results = WinklerResult(
        _name=f"{model.name} ({model.pile.name})",
        _d=disp_to_df(model, d),
        _f=structural_forces_to_df(model, q_int),
        _Q=reaction_forces_to_df(model, Q),
        _details={
            "converged @ iter no.": 1,
            "error": 0.0,
            "tolerance": None,
        },
    )

    return results


def winkler(model, max_iter: int = 100):
    """
    Function where loading or displacement defined in the model boundary conditions
    are used to solve the system of equations via the iterative Newton-Raphson scheme.

    Parameters
    ----------
    model : `openpile.construct.Model` object
        Model where structure and boundary conditions are defined.
    max_iter: int, by defaut 100
        maximum number of iterations for convergence

    Returns
    -------
    results : `openpile.analyses.Result` object
        Results of the analysis
    """

    if model.soil is None:
        UserWarning("A SoilProfile must be provided to the model before running this model.")

    else:
        # initialise global force
        F = kernel.mesh_to_global_force_dof_vector(model.global_forces)
        # initiliase prescribed displacement vector
        U = kernel.mesh_to_global_disp_dof_vector(model.global_disp)
        # initialise displacement vectors
        d = np.zeros(U.shape)
        # initialise global stiffness matrix
        K = kernel.build_stiffness_matrix(model, u=d, kind="initial")
        # initialise global supports vector
        supports = (kernel.mesh_to_global_restrained_dof_vector(model.global_restrained)) | (
            U != 0.0
        )

        # validate boundary conditions
        # validation.check_boundary_conditions(model)

        # Initialise residual forces
        Rg = deepcopy(F)

        # incremental calculations to convergence
        iter_no = 0
        while iter_no <= max_iter:

            # solve system
            try:
                u_inc, Q = kernel.solve_equations(K, Rg, U, restraints=supports)
                # bump iteration number
                iter_no += 1
            except np.linalg.LinAlgError:
                print(
                    """Cannot converge. Failure of the pile-soil system.\n
                      Boundary conditions may not be realistic or values may be too large."""
                )
                # dummy output vars
                Q = np.full(F.shape, np.nan)
                d = np.full(U.shape, np.nan)
                nr_tol = np.nan
                break

            # External forces
            F_ext = F - Q
            control = np.linalg.norm(F_ext)
            nr_tol = 1e-4 * control

            # add up increment displacements
            d += u_inc

            # calculate internal forces
            K_secant = kernel.build_stiffness_matrix(model, u=d, kind="secant")
            F_int = -K_secant.dot(d)

            # calculate residual forces
            Rg = F_ext + F_int

            # check if converged
            if np.linalg.norm(Rg[~supports]) < nr_tol and iter_no > 1:
                # do not accept convergence without iteration (without a second call to solve equations)
                print(f"Converged at iteration no. {iter_no}")

                # final stiffness matrix
                K_final = kernel.build_stiffness_matrix(model, u=d, kind="secant")

                # calculate final reaction forces
                _, Q = kernel.solve_equations(
                    K_final,
                    kernel.mesh_to_global_force_dof_vector(model.global_forces),
                    kernel.mesh_to_global_disp_dof_vector(model.global_disp),
                    restraints=supports,
                )

                break

            if iter_no == 100:
                print("Not converged after 100 iterations.")

                # dummy output vars
                Q = np.full(F.shape, np.nan)
                d = np.full(U.shape, np.nan)

            # re-calculate global stiffness matrix for next iterations
            K = kernel.build_stiffness_matrix(model, u=d, kind="tangent")

            # reset prescribed displacements to converge properly in case
            # of displacement-driven analysis
            U[:] = 0.0

        # Internal forces calculations with dim(nelem,6,6)
        q_int = kernel.pile_internal_forces(model, d)

        # Final results
        results = WinklerResult(
            _name=f"{model.name} ({model.pile.name}/{model.soil.name})",
            _d=disp_to_df(model, d),
            _f=structural_forces_to_df(model, q_int),
            _Q=reaction_forces_to_df(model, Q),
            _dist_mob=springs_mob_to_df(model, d),
            _hb_mob=(
                abs(d[-2])
                * kernel.calculate_base_spring_stiffness(d[-2], model._Hb_spring, kind="secant"),
                model._Hb_spring.flatten().max(),
            ),
            _mb_mob=(
                abs(d[-1])
                * kernel.calculate_base_spring_stiffness(d[-1], model._Mb_spring, kind="secant"),
                model._Mb_spring.flatten().max(),
            ),
            _details={
                "converged @ iter no.": iter_no,
                "error [kN]": round(np.linalg.norm(Rg[~supports]), 3),
                "tolerance [kN]": round(nr_tol, 3),
            },
        )

        return results


def simple_winkler_analysis(*args, **kwargs):
    """
    .. versionremoved: 1.0.0
        Use :func:`winkler` instead that keeps the same functional behaviour.
    """


def simple_beam_analysis(*args, **kwargs):
    """
    .. versionremoved: 1.0.0
        Use :func:`beam` instead that keeps the same functional behaviour.

    """



# import libraries
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt


from matplotlib.patches import FancyArrowPatch, Rectangle
from matplotlib.ticker import MultipleLocator, AutoMinorLocator
from openpile.core.misc import generate_color_string

mpl.rcParams["figure.subplot.wspace"] = 0.4


def plot_deflection(result):
    fig, ax = plt.subplots()

    fig.suptitle(f"{result._name} - Pile Deflection")

    ax = U_plot(ax, result)

    return fig


def plot_settlement(result):
    fig, (ax1, ax2) = plt.subplots(1, 2)

    fig.suptitle(f"{result._name} - Pile settlements")

    ax1 = settlement_plot(ax1, result)
    ax2 = F_plot(ax2, result, "N [kN]")

    ax2.set_yticklabels("")
    ax2.set_ylabel("")

    return fig


def plot_forces(result):
    # create 4 subplots with (deflectiom, normal force, shear force, bending moment)

    fig, (ax1, ax2, ax3) = plt.subplots(1, 3)

    fig.suptitle(f"{result._name} - Sectional forces")

    ax1 = F_plot(ax1, result, "N [kN]")
    ax2 = F_plot(ax2, result, "V [kN]")
    ax3 = F_plot(ax3, result, "M [kNm]")

    for axis in [ax2, ax3]:
        axis.set_yticklabels("")
        axis.set_ylabel("")

    return fig


def plot_results(result):
    # create 4 subplots with (deflectiom, normal force, shear force, bending moment)

    fig, (ax1, ax2, ax3, ax4) = plt.subplots(1, 4)

    fig.suptitle(f"{result._name} - Analysis results")

    ax1 = U_plot(ax1, result)
    ax2 = F_plot(ax2, result, "N [kN]")
    ax3 = F_plot(ax3, result, "V [kN]")
    ax4 = F_plot(ax4, result, "M [kNm]")

    for axis in [ax2, ax3, ax4]:
        axis.set_yticklabels("")
        axis.set_ylabel("")

    return fig


def soil_plot(SoilProfile, ax=None):
    def add_soil_profile(SoilProfile, ax, pile=None):

        ax.set_title(label=f"Soil Profile overview - {SoilProfile.name}")
        ax.set_xlim(left=0, right=1)

        # make data
        offset = 5
        yBot = SoilProfile.bottom_elevation
        if pile is None:
            yTop = max(SoilProfile.top_elevation + offset, SoilProfile.water_line + offset)
        else:
            yTop = max(
                SoilProfile.top_elevation + offset,
                SoilProfile.water_line + offset,
                pile.top_elevation + offset,
            )

        # axes
        ax.set_ylim(bottom=yBot, top=yTop)
        ax.set_ylabel("Elevation [m VREF]")
        ax.set_xticks([])

        for layer in SoilProfile.layers:
            ax.add_patch(
                Rectangle(
                    xy=(-100, layer.bottom),
                    width=200,
                    height=layer.top - layer.bottom,
                    facecolor=layer.color,
                )
            )

            ax.text(
                0.02,
                0.5 * (layer.top + layer.bottom),
                layer.name,
                bbox={"facecolor": [0.98, 0.96, 0.85], "alpha": 1, "edgecolor": "none", "pad": 1},
            )

        ax.plot(
            np.array([-100, 0.1, 100]),
            SoilProfile.water_line + np.zeros((3)),
            mfc="dodgerblue",
            marker=7,
            linewidth=1,
            color="dodgerblue",
        )

        # grid
        ax.minorticks_on()
        ax.grid()
        ax.grid(axis="y", which="minor", color=[0.75, 0.75, 0.75], linestyle="-", linewidth=0.5)

        ax.plot(
            np.array([-100, 0.1, 100]),
            SoilProfile.water_line + np.zeros((3)),
            mfc="dodgerblue",
            marker=7,
            linewidth=1,
            color="dodgerblue",
        )

        return ax

    if ax is None:
        _, ax = plt.subplots()
    ax = add_soil_profile(SoilProfile, ax, pile=None)

    return ax


def connectivity_plot(model, ax=None):
    # TODO docstring

    support_color = "b"
    # create 4 subplots with (deflectiom, normal force, shear force, bending moment)
    if ax is None:
        fig, ax = plt.subplots()
    ax.set_ylabel("z [m]")
    ax.set_xlabel("y [m]")
    ax.set_title(f"{model.name} - Connectivity plot")
    ax.axis("equal")
    ax.grid(which="both")

    # plot mesh with + scatter points to see nodes.
    z = model.nodes_coordinates["z [m]"]
    y = model.nodes_coordinates["y [m]"]
    ax.plot(y, z, "-k", marker="+")

    total_length = (
        (model.nodes_coordinates["z [m]"].max() - model.nodes_coordinates["z [m]"].min()) ** 2
        + (model.nodes_coordinates["y [m]"].max() - model.nodes_coordinates["y [m]"].min()) ** 2
    ) ** (0.5)

    ylim = ax.get_ylim()

    # plots SUPPORTS
    # Plot supports along z
    support_along_z = model.global_restrained["Tz"].values
    support_along_z_down = np.copy(support_along_z)
    support_along_z_down[-1] = False
    support_along_z_up = np.copy(support_along_z)
    support_along_z_up[:-1] = False
    ax.scatter(
        y[support_along_z_down],
        z[support_along_z_down],
        color=support_color,
        marker=7,
        s=100,
    )
    ax.scatter(
        y[support_along_z_up],
        z[support_along_z_up],
        color=support_color,
        marker=6,
        s=100,
    )

    # Plot supports along y
    support_along_y = model.global_restrained["Ty"].values
    ax.scatter(y[support_along_y], z[support_along_y], color=support_color, marker=5, s=100)

    # Plot supports along z
    support_along_z = model.global_restrained["Rx"].values
    ax.scatter(y[support_along_z], z[support_along_z], color=support_color, marker="s", s=35)

    # plot LOADS
    arrows = []

    normalized_arrow_size = (
        0.10 * total_length
    )  # max arrow length will be 20% of the total structure length

    load_max = model.global_forces["Py [kN]"].abs().max()
    for yval, zval, load in zip(z, y, model.global_forces["Py [kN]"]):
        if load == 0:
            pass
        else:
            style = "Simple, tail_width=1, head_width=5, head_length=3"
            kw = dict(arrowstyle=style, color="r")
            arrow_length = normalized_arrow_size * abs(load / load_max)
            if load > 0:
                arrows.append(FancyArrowPatch((-arrow_length, yval), (zval, yval), **kw))
            elif load < 0:
                arrows.append(FancyArrowPatch((arrow_length, yval), (zval, yval), **kw))

    load_max = model.global_forces["Pz [kN]"].abs().max()
    for idx, (yval, zval, load) in enumerate(zip(z, y, model.global_forces["Pz [kN]"])):
        if load == 0:
            pass
        else:
            style = "Simple, tail_width=1, head_width=5, head_length=3"
            kw = dict(arrowstyle=style, color="r")
            arrow_length = normalized_arrow_size * abs(load / load_max)
            if load > 0:
                if idx == len(z) - 1:
                    arrows.append(FancyArrowPatch((zval, yval), (zval, yval + arrow_length), **kw))
                else:
                    arrows.append(FancyArrowPatch((zval, yval - arrow_length), (zval, yval), **kw))
            elif load < 0:
                if idx == len(z) - 1:
                    arrows.append(FancyArrowPatch((zval, yval), (zval, yval - arrow_length), **kw))
                else:
                    arrows.append(FancyArrowPatch((zval, yval + arrow_length), (zval, yval), **kw))

    load_max = model.global_forces["Mx [kNm]"].abs().max()
    for idx, (yval, zval, load) in enumerate(zip(z, y, model.global_forces["Mx [kNm]"])):
        if load == 0:
            pass
        else:
            kw = dict(arrowstyle=style, color="r")
            arrow_length = normalized_arrow_size * abs(load / load_max)
            style = "Simple, tail_width=1, head_width=5, head_length=3"
            if load > 0:
                if idx == len(z) - 1:
                    arrows.append(
                        FancyArrowPatch(
                            (arrow_length / 1.5, yval),
                            (-arrow_length / 1.5, yval),
                            connectionstyle="arc3,rad=0.5",
                            **kw,
                        )
                    )
                else:
                    arrows.append(
                        FancyArrowPatch(
                            (-arrow_length / 1.5, yval),
                            (arrow_length / 1.5, yval),
                            connectionstyle="arc3,rad=0.5",
                            **kw,
                        )
                    )
            elif load < 0:
                if idx == len(z) - 1:
                    arrows.append(
                        FancyArrowPatch(
                            (arrow_length / 1.5, yval),
                            (-arrow_length / 1.5, yval),
                            connectionstyle="arc3,rad=-0.5",
                            **kw,
                        )
                    )
                else:
                    arrows.append(
                        FancyArrowPatch(
                            (-arrow_length / 1.5, yval),
                            (arrow_length / 1.5, yval),
                            connectionstyle="arc3,rad=-0.5",
                            **kw,
                        )
                    )

    ax.set_ylim(ylim[0] - 0.11 * total_length, ylim[1] + 0.11 * total_length)

    if model.soil is not None:

        ax.set_ylim(
            min(model.bottom, ylim[0]) - 0.11 * total_length,
            max(model.top, ylim[1]) + 0.11 * total_length,
        )

        for layer in model.soil.layers:
            ax.add_patch(
                Rectangle(
                    xy=(-2 * total_length, layer.bottom),
                    width=4 * total_length,
                    height=layer.top - layer.bottom,
                    facecolor=layer.color,
                    alpha=0.4,
                )
            )

            ax.text(
                -1 * total_length,
                0.5 * (layer.top + layer.bottom) + 0.1 * (layer.top - layer.bottom),
                layer.name,
                bbox={"facecolor": [0.98, 0.96, 0.85], "alpha": 1, "edgecolor": "none", "pad": 1},
            )

            ax.plot(
                np.array([-2 * total_length, -0.6 * total_length, 2 * total_length]),
                model.soil.water_line + np.zeros((3)),
                mfc="dodgerblue",
                marker=7,
                linewidth=1,
                color="dodgerblue",
            )
        ax.set_xlim((-0.7 * total_length, 0.3 * total_length))

    else:
        ax.set_xlim((-0.5 * total_length, 0.5 * total_length))

    for arrow in arrows:
        ax.add_patch(arrow)

    return ax


def U_plot(axis: plt.axis, result):
    # TODO docstring

    axis.set_ylabel("Elevation [m VREF]", fontsize=8)
    axis.set_xlabel("Deflection [mm]", fontsize=8)
    axis.tick_params(axis="both", labelsize=8)
    axis.grid(which="both")

    y = result.displacements["Elevation [m]"].values
    x = np.zeros(shape=y.shape)
    deflection = result.displacements["Deflection [m]"] * 1e3

    axis.plot(x, y, color="0.4")
    axis.plot(deflection, y, color="0.0", lw=2)

    return axis


def settlement_plot(axis: plt.axis, result):
    # TODO docstring

    axis.set_ylabel("Elevation [m VREF]", fontsize=8)
    axis.set_xlabel("Settlement [mm]", fontsize=8)
    axis.tick_params(axis="both", labelsize=8)
    axis.grid(which="both")

    y = result.displacements["Elevation [m]"].values
    x = np.zeros(shape=y.shape)
    settlement = result.settlement["Settlement [m]"] * 1e3

    axis.plot(x, y, color="0.4")
    axis.plot(settlement, y, color="0.0", lw=2)

    return axis


def F_plot(axis: plt.axis, result, force: str):
    # TODO docstring

    # Define plot colors
    force_facecolor = "#E6DAA6"  # beige
    force_edgecolor = "#AAA662"  # khaki

    axis.set_ylabel("Elevation [m VREF]", fontsize=8)
    axis.set_xlabel(force, fontsize=8)
    axis.tick_params(axis="both", labelsize=8)
    axis.grid(which="both")

    f = result.forces[force]
    y = result.forces["Elevation [m]"]

    axis.fill_betweenx(y, f, edgecolor=force_edgecolor, facecolor=force_facecolor)
    axis.plot(np.zeros(shape=y.shape), y, color="0.4")

    axis.set_xlim(
        [
            min(0, f.min() - 0.1 * abs(f.min() + 1.0)),
            max(0, f.max() + 0.1 * abs(f.max() + 1.0)),
        ]
    )

    return axis



import pandas as pd
import numpy as np


def validate_bc(bc_list, bc_cls):
    """
    helper function to validate boundary condition
    """
    err = False
    flag = []
    # check if bcs overlap
    bc_checklist = set([(bc.elevation, bc.x) for bc in bc_list if isinstance(bc, bc_cls)])
    if len(bc_checklist) > 1:
        if len(set([bc[0] for bc in bc_checklist])) != len([bc[0] for bc in bc_checklist]):
            flag.append("x")
            err = True
    bc_checklist = set([(bc.elevation, bc.y) for bc in bc_list if isinstance(bc, bc_cls)])
    if len(bc_checklist) > 1:
        if len(set([bc[0] for bc in bc_checklist])) != len([bc[0] for bc in bc_checklist]):
            flag.append("y")
            err = True
    bc_checklist = set([(bc.elevation, bc.z) for bc in bc_list if isinstance(bc, bc_cls)])
    if len(bc_checklist) > 1:
        if len(set([bc[0] for bc in bc_checklist])) != len([bc[0] for bc in bc_checklist]):
            flag.append("z")
            err = True
    if err:
        raise ValueError(
            f"Multiple boundary conditions ({bc_cls.__name__}) are given along {', and '.join(flag)}. axis and at same elevation."
        )


def apply_bc(nodes_elevations, zglobal, yglobal, xglobal, bc_list, bc_cls, output_text):
    """
    helper function to apply boundary condition
    """
    # apply boundary condition
    for bc in bc_list:
        if isinstance(bc, bc_cls):
            check = np.isclose(
                nodes_elevations, np.tile(bc.elevation, nodes_elevations.shape), atol=0.001
            )
            if any(check):
                # one node correspond, extract node
                node_idx = next((i for i, x in enumerate(check) if x == True), None)
                # apply loads at this node
                if bc.x:
                    xglobal[node_idx] = bc.x
                if bc.y:
                    yglobal[node_idx] = bc.y
                if bc.z:
                    zglobal[node_idx] = bc.z
            else:
                if bc.elevation > nodes_elevations[0] or bc.elevation < nodes_elevations[-1]:
                    print(
                        f"{output_text} not applied! The chosen elevation is outside the mesh. The {output_text} must be applied on the structure."
                    )
                else:
                    print(
                        f"{output_text} not applied! The chosen elevation is not meshed as a node. Please include elevation in `x2mesh` variable when creating the Model."
                    )
    return zglobal, yglobal, xglobal


def parameter2elements(objects: list, key: callable, elem_top: list, elem_bottom: list):
    """converts a list of pile sections into a list of elements

    Objects must be either a list of pile sections (openpile.construct.PileSection) or a list of soil layers (openpile.construct.Layer)

    """

    elem_top = np.array(elem_top)
    elem_bottom = np.array(elem_bottom)
    # create a NaN array with same array as elem_z_top
    elem_param = np.full(elem_top.size, np.nan)

    for obj in objects:
        top_limit, bottom_limit = obj.top, obj.bottom
        idx = np.where((elem_top <= top_limit) & (elem_bottom >= bottom_limit))[0]
        elem_param[idx] = key(obj)

    return elem_param


def get_tip_sig_v_eff(
    tip_elevation: float,
    water_elevation: float,
    layers: list,  # List[openpile.construct.Layer]
    sig_v_mudline: float = 0,
    water_unit_weight: float = 10.0,
):
    """Calculates the effective vertical stress at the pile tip"""
    sig_v_tip = sig_v_mudline
    water_unit_weight = 10.0

    for layer in sorted(layers, key=lambda x: -x.top):
        buoyant_weight = layer.weight - water_unit_weight
        if tip_elevation <= layer.bottom:
            if water_elevation <= layer.bottom:
                w = layer.weight
            elif water_elevation < layer.top:
                w = (
                    layer.weight * (layer.top - water_elevation)
                    + buoyant_weight * (water_elevation - layer.bottom)
                ) / (layer.top - layer.bottom)
            else:
                w = buoyant_weight

            sig_v_tip += w * (layer.top - layer.bottom)

        elif tip_elevation < layer.top and water_elevation > layer.bottom:
            if water_elevation <= tip_elevation:
                w = layer.weight
            elif water_elevation < layer.top:
                w = (
                    layer.weight * (layer.top - water_elevation)
                    + buoyant_weight * (water_elevation - tip_elevation)
                ) / (layer.top - tip_elevation)
            else:
                w = buoyant_weight

            sig_v_tip += w * (layer.top - tip_elevation)

        return sig_v_tip


def check_springs(arr):
    check_nan = np.isnan(arr).any()
    check_negative = (arr < 0).any()

    return check_nan or check_negative


def get_coordinates(pile, soil, x2mesh, coarseness) -> pd.DataFrame:
    # Primary discretisation over z-axis
    z = np.array([], dtype=np.float16)
    # add get pile relevant sections
    z = np.append(z, pile.data["Elevation [m]"].values)
    # add soil relevant layers and others
    if soil is not None:
        soil_elevations = np.array(
            [x.top for x in soil.layers] + [x.bottom for x in soil.layers],
            dtype=float,
        )
        if any(soil_elevations < pile.bottom_elevation):
            soil_elevations = np.append(pile.bottom_elevation, soil_elevations)
            soil_elevations = soil_elevations[soil_elevations >= pile.bottom_elevation]
        z = np.append(z, soil_elevations)
    # add user-defined elevation
    if x2mesh is None:
        x2mesh = []
    z = np.append(z, x2mesh)

    # get unique values and sort in reverse order
    z = np.unique(z)[::-1]

    # Secondary discretisation over z-axis depending on coarseness factor
    z_secondary = np.array([], dtype=np.float16)
    for i in range(len(z) - 1):
        spacing = z[i] - z[i + 1]
        new_spacing = spacing
        divider = 1
        while new_spacing > coarseness:
            divider += 1
            new_spacing = spacing / divider
        new_z = z[i] - (np.arange(start=1, stop=divider) * np.tile(new_spacing, (divider - 1)))
        z_secondary = np.append(z_secondary, new_z)

    # assemble x- coordinates
    z = np.append(z, z_secondary)
    z = np.unique(z)[::-1]

    # dummy y- coordinates
    y = np.zeros(shape=z.shape)

    # create dataframe coordinates
    nodes = pd.DataFrame(
        data={
            "z [m]": z,
            "y [m]": y,
        },
        dtype=float,
    ).round(4)
    nodes.index.name = "Node no."

    element = pd.DataFrame(
        data={
            "z_top [m]": z[:-1],
            "z_bottom [m]": z[1:],
            "y_top [m]": y[:-1],
            "y_bottom [m]": y[1:],
        },
        dtype=float,
    ).round(4)
    element.index.name = "Element no."

    return nodes, element


def get_soil_profile(soil) -> pd.DataFrame:
    top_elevations = [x.top for x in soil.layers]
    bottom_elevations = [x.bottom for x in soil.layers]
    soil_weights = [x.weight for x in soil.layers]

    idx_sort = np.argsort(top_elevations)[::-1]

    top_elevations = [top_elevations[i] for i in idx_sort]
    soil_weights = [soil_weights[i] for i in idx_sort]
    bottom_elevations = [bottom_elevations[i] for i in idx_sort]

    # elevation in model w.r.t to x axis
    x = top_elevations

    return pd.DataFrame(
        data={"Top soil layer [m]": x, "Unit Weight [kN/m3]": soil_weights},
        dtype=np.float64,
    )


def get_soil_properties(pile, soil, x2mesh, coarseness):
    # dummy allocation
    soil_properties = None
    # create soil properties
    if soil is not None:
        element_coordinates = get_coordinates(pile, soil, x2mesh, coarseness)[1]
        soil_properties = pd.merge_asof(
            left=element_coordinates[["z_top [m]", "z_bottom [m]"]].sort_values(by=["z_top [m]"]),
            right=get_soil_profile(soil).sort_values(by=["Top soil layer [m]"]),
            left_on="z_top [m]",
            right_on="Top soil layer [m]",
            direction="forward",
        ).sort_values(by=["z_top [m]"], ascending=False)
        # add elevation of element w.r.t. ground level
        soil_properties["zg_top [m]"] = soil_properties["z_top [m]"] - soil.top_elevation
        soil_properties["zg_bottom [m]"] = soil_properties["z_bottom [m]"] - soil.top_elevation
        # add vertical stress at top and bottom of each element
        condition_below_water_table = soil_properties["z_top [m]"] <= soil.water_line
        soil_properties.loc[condition_below_water_table, "Unit Weight [kN/m3]"] = (
            soil_properties["Unit Weight [kN/m3]"][condition_below_water_table] - 10.0
        )
        s = (soil_properties["z_top [m]"] - soil_properties["z_bottom [m]"]) * soil_properties[
            "Unit Weight [kN/m3]"
        ]
        soil_properties["sigma_v top [kPa]"] = np.insert(
            s.cumsum().values[:-1],
            np.where(soil_properties["z_top [m]"].values == soil.top_elevation)[0],
            0.0,
        )
        soil_properties["sigma_v bottom [kPa]"] = s.cumsum()
        # reset index
        soil_properties.reset_index(inplace=True, drop=True)

        return soil_properties



"""
`py_curves` module
------------------

"""

# Import libraries
import math as m
import numpy as np
from numba import njit, prange
from random import random

from openpile.core.misc import conic

# SPRING FUNCTIONS --------------------------------------------
@njit(cache=True)
def bothkennar_clay(
    X: float,
    Su: float,
    G0: float,
    D: float,
    output_length: int = 20,
):
    """
    Creates a spring from the PISA clay formulation
    published by Burd et al 2020 (see [BABH20]_) and calibrated based on Bothkennar clay
    response (a normally consolidated soft clay).

    Parameters
    ----------
    X : float
        Depth below ground level [unit: m]
    Su : float
        Undrained shear strength [unit: kPa]
    G0 : float
        Small-strain shear modulus [unit: kPa]
    D : float
        Pile diameter [unit: m]
    output_length : int, optional
        Number of datapoints in the curve, by default 20

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]

    See also
    --------
    :py:func:`openpile.utils.py_curves.cowden_clay`, :py:func:`openpile.utils.py_curves.custom_clay`
    """

    # # Bothkennar clay parameters
    v_pu = 173.8
    k_p1 = 12.05
    k_p2 = -1.547
    n_p1 = 0.7204
    n_p2 = -0.002679
    p_u1 = 7.743
    p_u2 = -3.945

    # Depth variation parameters
    v_max = v_pu
    k = k_p1 + k_p2 * X / D
    n = n_p1 + n_p2 * X / D
    p_max = p_u1 + p_u2 * m.exp(-0.8456 * X / D)

    # calculate normsalised conic function
    y, p = conic(v_max, n, k, p_max, output_length)

    # return non-normalised curve
    return y * (Su * D / G0), p * (Su * D)


@njit(cache=True)
def cowden_clay(
    X: float,
    Su: float,
    G0: float,
    D: float,
    output_length: int = 20,
):
    """
    Creates a spring from the PISA clay formulation
    published by Byrne et al 2020 (see [BHBG20]_) and calibrated based pile
    load tests at Cowden (north east coast of England).

    Parameters
    ----------
    X : float
        Depth below ground level [unit: m]
    Su : float
        Undrained shear strength [unit: kPa]
    G0 : float
        Small-strain shear modulus [unit: kPa]
    D : float
        Pile diameter [unit: m]
    output_length : int, optional
        Number of datapoints in the curve, by default 20

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]

    See also
    --------
    :py:func:`openpile.utils.py_curves.bothkennar_clay`, :py:func:`openpile.utils.py_curves.custom_clay`
    """

    # # Cowden clay parameters
    v_pu = 241.4
    k_p1 = 10.6
    k_p2 = -1.650
    n_p1 = 0.9390
    n_p2 = -0.03345
    p_u1 = 10.7
    p_u2 = -7.101

    # Depth variation parameters
    v_max = v_pu
    k = k_p1 + k_p2 * X / D
    n = n_p1 + n_p2 * X / D
    p_max = p_u1 + p_u2 * m.exp(-0.3085 * X / D)

    # calculate normsalised conic function
    y, p = conic(v_max, n, k, p_max, output_length)

    # return non-normalised curve
    return y * (Su * D / G0), p * (Su * D)


@njit(cache=True)
def dunkirk_sand(
    sig: float,
    X: float,
    Dr: float,
    G0: float,
    D: float,
    L: float,
    output_length: int = 20,
):
    r"""
    Creates a lateral spring from the PISA sand formulation
    published  by Burd et al (2020) (see [BTZA20]_).
    Also called the General Dunkirk Sand Model (GDSM).

    Parameters
    ----------
    sig : float
        vertical/overburden effective stress [unit: kPa]
    X : float
        Depth below ground level [unit: m]
    Dr : float
        Sand relative density Value must be between 0 and 100 [unit: -]
    G0 : float
        Small-strain shear modulus [unit: kPa]
    D : float
        Pile diameter [unit: m]
    L : float
        Embedded pile length [unit: m]
    output_length : int, optional
        Number of datapoints in the curve, by default 20

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]

    Notes
    -----

    The curve backbone is defined as a conic function, see below.

    .. figure:: _static/PISA_conic_function.png
        :width: 80%

        PISA Conic function: (a) conic form; (b) bilinear form, after [BHBG20]_.

    """
    output_length = max(8, output_length)

    # correct relative density for decimal value
    Dr = Dr / 100

    # Generalised Dunkirk Sand Model parameters
    v_pu = 146.1 - 92.11 * Dr
    k_p1 = 8.731 - 0.6982 * Dr
    k_p2 = -0.9178
    n_p = 0.917 + 0.06193 * Dr
    p_u1 = 0.3667 + 25.89 * Dr
    p_u2 = 0.3375 - 8.9 * Dr

    # Depth variation parameters
    v_max = v_pu
    k = k_p1 + k_p2 * X / D
    n = n_p
    p_max = p_u1 + p_u2 * X / L

    # calculate normsalised conic function
    y, p = conic(v_max, n, k, p_max, output_length)

    # return non-normalised curve
    return y * (sig * D / G0), p * (sig * D)


# API sand function
@njit(parallel=True, cache=True)
def api_sand(
    sig: float,
    X: float,
    phi: float,
    D: float,
    kind: str = "static",
    below_water_table: bool = True,
    k: float = 0.0,
    ymax: float = 0.0,
    output_length: int = 20,
):
    r"""Creates the API sand p-y curve from relevant input.

    
    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    X: float
        Depth of the curve w.r.t. mudline [unit: m]
    phi: float
        internal angle of friction of the sand layer [unit: degrees]
    D: float
        Pile width [unit: m]
    kind: str, by default "static"
        types of curves, can be of ("static","cyclic")
    below_water_table: bool, by default False
        switch to calculate initial subgrade modulus below/above water table
    k: float, by default 0.0
        user-defined initial subgrade modulus [kN/m^3], if kept as zero, it is calculated as per API guidelines, see Notes below
    ymax: float, by default 0.0
        maximum value of y, default goes to 99.9% of ultimate resistance
    output_length: int, by default 20
        Number of discrete point along the springs, cannot be less than 8

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]

        
    Notes
    -----
    
    p-y formulation
        The API sand **p-y formulation** is presented in both the API and DNVGL standards,
        see, [DNV-RP-C212]_, [API2000]_ or [API2014]_.

        Granular soils are modelled by the sand p-y model as described 
        with the following backbone formula:
        
        .. math::

            p = A \cdot P_{max} \cdot \tanh \left( \frac{k \cdot X}{A \cdot P_{max} }  y \right) 

        where:

        * :math:`A` is a factor to account for static of cyclic loading 
        * :math:`P_{max}` is the ultimate resistance of the p-y curve 
        * :math:`k` is the initial modulus of subgrade reaction
        * :math:`X` is the depth below mudline of the p-y curve.

    Factor A
        The factor A takes into account whether the curve represent 
        static(also called monotonic) or cycling loading and is equal to:

        .. math::

            A = 
            \begin{cases} 
            \begin{split}
            0.9 & \text{  for cyclic loading} \\ 
            \\
            3 - 0.8 \frac{X}{D} \ge 0.9 & \text{  for static loading}
                \end{split}
            \end{cases}

        where:

        * :math:`D` is the pile diameter. 
    
    Initial subgrade reaction modulus
        The initial subgrade reaction, represented by the factor k is
        approximated by the following equation in which the output is given in kN/m³ 
        and where :math:`\phi` is inserted in degrees: 

        .. math::

            k = 
            \begin{cases} 
            \begin{split}
            197.8 \cdot \phi^2 - 10232 \cdot \phi + 136820 \ge 5400 & \text{ ,  below water table} \\ 
            \\
            215.3 \cdot \phi^2 - 8232 \cdot \phi + 63657 \ge 5400  & \text{ ,  above water table}
            \end{split}
            \end{cases}

        The equation is a fit to the recommended values in [API2000]_.  The correspondence 
        of this fit is illustrated in below figure:

        .. figure:: /_static/py_API_sand/k_vs_phi.jpg
            :width: 80%

            Subgrade reaction moduli fits calculated by openpile.

        .. note::

            The initial subgrade modulus can be user-defined by setting the parameter 'k' to a value greater than zero.
            Furthermore, many researchers have proposed different values for the initial subgrade modulus, see :py:class:`openpile.soilmodels.hooks.InitialSubgradeReaction`.

    Ultimate resistance
        The ultimate resistance :math:`P_{max}` is calculated via the coefficients C1, C2 and C3 found 
        in the below figure. 

        .. figure:: _static/py_API_sand/C_coeffs_graph.jpg
            :width: 80%

            Coefficients to calculate the maximum resistance. (as given in [MuOn84]_) 

        The ultimate resistance is found via the below equation:

        .. math::

            P_{max} = \left( 
                C1 \cdot \sigma^{\prime} \cdot X + C2 \cdot \sigma^{\prime} \cdot D \right) \lt
                C3 \cdot \sigma^{\prime} \cdot D 

        where:

        * :math:`\sigma^{\prime}` is the vertical effective stress

    """
    output_length = max(8, output_length)

    # A value - only thing that changes between cyclic or static
    if kind == "static":
        A = max(0.9, 3 - 0.8 * X / D)
    else:
        A = 0.9

    # initial subgrade modulus in kpa from fit with API (2014)
    if k != 0:
        if k < 0:
            raise ValueError("'initial_subgrade_modulus' must be stricly positive.")
        k_phi = k
    else:
        if below_water_table:
            k_phi = max((0.1978 * phi**2 - 10.232 * phi + 136.82) * 1000, 5400)
        else:
            k_phi = max((0.2153 * phi**2 - 8.232 * phi + 63.657) * 1000, 5400)

    # Calculate Pmax (regular API)
    ## Factors according to Mosher and Dawkins 2008.(regular API)
    b = 0.4
    Beta = 45 + phi / 2
    rad = m.pi / 180
    C1 = (
        (b * m.tan(phi * rad) * m.sin(Beta * rad))
        / (m.tan((Beta - phi) * rad) * m.cos((phi / 2) * rad))
        + ((m.tan(Beta * rad)) ** 2 * m.tan((phi / 2) * rad)) / (m.tan((Beta - phi) * rad))
        + b * m.tan(Beta * rad) * (m.tan(phi * rad) * m.sin(Beta * rad) - m.tan((phi / 2) * rad))
    )
    C2 = m.tan(Beta * rad) / m.tan((Beta - phi) * rad) - (m.tan((45 - phi / 2) * rad)) ** 2
    C3 = b * m.tan(phi * rad) * (m.tan(Beta * rad)) ** 4 + (m.tan((45 - phi / 2) * rad)) ** 2 * (
        (m.tan(Beta * rad)) ** 8 - 1
    )

    ## Pmax for shallow and deep zones (regular API)
    Pmax = min(C3 * sig * D, C1 * sig * X + C2 * sig * D)

    # creation of 'y' array
    if ymax == 0.0:
        # ensure X cannot be zero
        z = max(X, 0.1)
        if Pmax == 0:
            f = (C3 * D + C2 * D) / 2
        else:
            f = Pmax
        ymax = 4 * A * f / (k_phi * z)

    # determine y vector from 0 to ymax
    # y_ini = np.array([0,],dtype=np.float64)
    # y = np.append(y_ini,np.logspace(np.log10(0.0001*D),np.log10(ymax),output_length-1))
    y = np.linspace(0, ymax, output_length)

    ## calculate p vector
    p = np.zeros(shape=len(y), dtype=np.float32)
    for i in prange(len(y)):
        if Pmax == 0:
            p[i] = 0
        else:
            p[i] = A * Pmax * m.tanh((k_phi * X * y[i]) / (A * Pmax))

    return y, p


# API clay function
@njit(parallel=True, cache=True)
def api_clay(
    sig: float,
    X: float,
    Su: float,
    eps50: float,
    D: float,
    J: float = 0.5,
    kind: str = "static",
    ymax: float = 0.0,
    output_length: int = 20,
):
    r"""
    Creates the API clay p-y curve from API RP2GEO (2014) from relevant input.

    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    X: float
        Depth of the curve w.r.t. mudline [unit: m]
    Su : float
        Undrained shear strength [unit: kPa]
    eps50: float
        strain at 50% ultimate resistance [-]
    D: float
        Pile width [unit: m]
    J: float, by default 0.5
        empirical factor varying depending on clay stiffness
    kind: str, by default "static"
        types of curves, can be of ("static","cyclic")
    ymax: float, by default 0.0
        maximum value of y, if null the maximum is calculated such that the whole curve is computed
    output_length: int, by default 20
        Number of discrete point along the springs, cannot be less than 8

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]

    See also
    --------

    :py:class:`openpile.soilmodels.API_clay`, :py:func:`openpile.utils.py_curves.matlock_1970`

    Notes
    -----

    The p-y clay formulation is presented in both the API and DNVGL standards,
    see [DNV-RP-C212]_ and [API2014]_.

    The p-y curve is defined as per Matlock original equations (see notes in :py:func:`openpile.utils.py_curves.matlock_1970`) but defined at specific points, in a piece-wise manner.

    Static piece-wise points are defined as follows:
        +------------------+-------------------+
        | :math:`p/P_{max}`|  :math:`y/y_{50}` |
        +==================+===================+
        | 0.00             |  0.0              |
        +------------------+-------------------+
        | 0.23             |  0.1              |
        +------------------+-------------------+
        | 0.33             |  0.3              |
        +------------------+-------------------+
        | 0.50             |  1.0              |
        +------------------+-------------------+
        | 0.72             |  3                |
        +------------------+-------------------+
        | 1.00             |  8                |
        +------------------+-------------------+
        | 1.00             |  :math:`\infty`   |
        +------------------+-------------------+

    Cyclic piece-wise points are defined as follows:
        +----------------------------------------------------+------------------+
        |                 :math:`p/P_{max}`                  | :math:`y/y_{50}` |
        +====================================================+==================+
        | 0.00                                               | 0.0              |
        +----------------------------------------------------+------------------+
        | 0.23                                               | 0.1              |
        +----------------------------------------------------+------------------+
        | 0.33                                               | 0.3              |
        +----------------------------------------------------+------------------+
        | 0.50                                               | 1.0              |
        +----------------------------------------------------+------------------+
        | 0.72                                               | 3.0              |
        +----------------------------------------------------+------------------+
        | :math:`\min\left(0.72; 0.72 \dfrac{z}{X_R}\right)` | 15.0             |
        +----------------------------------------------------+------------------+
        | :math:`\min\left(0.72; 0.72 \dfrac{z}{X_R}\right)` | :math:`\infty`   |
        +----------------------------------------------------+------------------+

        where:

        * :math:`z` is the depth below mudline
        * :math:`X_R` is the transition zone depth below mudline
        * :math:`y_{50}` is the equivalent displacement for deformation at 50% ultimate resistance
        * :math:`P_{max}` is the ultimate resistance of the p-y curve
    """

    output_length = max(8, output_length)

    # important variables
    y50 = 2.5 * eps50 * D
    if X == 0.0 or Su == 0:
        Xr = 2.5 * D
    else:
        Xr = max((6 * D) / (sig / X * D / Su + J), 2.5 * D)

    # creation of 'y' array
    if ymax == 0.0:
        ymax = 16 * y50

    # Calculate Pmax (regular API)
    ## Pmax for shallow and deep zones (regular API)
    Pmax_shallow = (3 * Su + sig) * D + J * Su * X
    Pmax_deep = 9 * Su * D
    Pmax = min(Pmax_deep, Pmax_shallow)

    ylist_in = [0.0, 0.1 * y50, 0.3 * y50, 1 * y50, 3 * y50, 8 * y50, 15 * y50, ymax]
    ylist_out = []
    for i in range(len(ylist_in)):
        if ylist_in[i] <= ymax:
            ylist_out.append(ylist_in[i])

    # determine y vector from 0 to ymax
    y = np.array(ylist_out, dtype=np.float32)
    add_values = output_length - len(y)
    add_y_values = []
    for _ in range(add_values):
        add_y_values.append(15 * y50 + random() * (ymax - 15 * y50))
    y = np.append(y, add_y_values)
    y = np.sort(y)

    # define p vector
    p = np.zeros(shape=len(y), dtype=np.float32)

    for i in prange(len(y)):
        if kind == "static":
            # derive static curve
            if y[i] > 8 * y50:
                p[i] = Pmax
            else:
                p[i] = 0.5 * Pmax * (y[i] / y50) ** 0.33
        else:
            # derive cyclic curve
            if X <= Xr:
                if y[i] > 15 * y50:
                    p[i] = 0.7185 * Pmax * X / Xr
                elif y[i] > 3 * y50:
                    p[i] = 0.7185 * Pmax * (1 - (1 - X / Xr) * (y[i] - 3 * y50) / (12 * y50))
                else:
                    p[i] = 0.5 * Pmax * (y[i] / y50) ** 0.33

            elif X > Xr:
                if y[i] > 3 * y50:
                    p[i] = 0.7185 * Pmax
                else:
                    p[i] = 0.5 * Pmax * (y[i] / y50) ** 0.33

    return y, p


# API clay function
@njit(parallel=True, cache=True)
def matlock_1970(
    sig: float,
    X: float,
    Su: float,
    eps50: float,
    D: float,
    J: float = 0.5,
    kind: str = "static",
    ymax: float = 0.0,
    output_length: int = 20,
):
    r"""
    Creates the original clay p-y curve from the work performed by [Matl70]_.

    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    X: float
        Depth of the curve w.r.t. mudline [unit: m]
    Su : float
        Undrained shear strength [unit: kPa]
    eps50: float
        strain at 50% ultimate resistance [-]
    D: float
        Pile width [unit: m]
    J: float, by default 0.5
        empirical factor varying depending on clay stiffness
    kind: str, by default "static"
        types of curves, can be of ("static","cyclic")
    ymax: float, by default 0.0
        maximum value of y, if null the maximum is calculated such that the whole curve is computed
    output_length: int, by default 20
        Number of discrete point along the springs, cannot be less than 9

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]

    See also
    --------
    :py:func:`openpile.utils.py_curves.api_clay`, :py:class:`openpile.soilmodels.API_clay`

    Notes
    -----

    Ultimate resistance
        The **ultimate resistance** is calculated via the capacity of two failure mechanisms,
        one that is shallow (wedge-type failure) and another that is deep (flow-around failure).

        .. math::

            P_{max} &= min(P_{shallow}, P_{deep})
            \\\\
            P_{shallow} &= D (3 S_u + \sigma^{\prime}) + J \cdot S_u \cdot X
            \\\\
            P_{deep} &=  9 \cdot S_u \cdot D

        where: 

        * :math:`S_u` is the undrained shear strßength in Unconfined and 
          unconsolidated (UU) Trixial tests.
        * :math:`\sigma^{\prime}` is the vertical effective stress.
        * :math:`J` is an empirical factor determined by Matlock to fit results 
          to pile load tests. This value can vary from 0.25 to 0.50 depending on 
          the clay characteristics
        * :math:`X` is the depth below ground level

        
    Strain normalization
        Strain normalization is performed with a parameter :math:`y_{50}` that is used to scale the curve with respect
        to the structure's scale and soil type.

        .. math::

            y_{50} = 2.5 \cdot \varepsilon_{50} \cdot D

        where: 

        * :math:`D` is the pile width or diameter
        * :math:`\varepsilon_{50}` is the strain at 50% ultimate resistance
          in Unconfined and unconsolidated (UU) Triaxial tests.

          
    Transition zone
        A transition zone is defined which corresponds to the depth at which the failure 
        around the pile is not governed by the free-field boundary, i.e. the ground level.
        Below the transition zone, a flow-around type of failure.

        The transition zone is defined by the following formula:

        .. math::

            X_R = \left( \frac{6 \cdot D}{\gamma^{\prime} \cdot \frac{D}{S_u} + J} \right) \ge  2.5 \cdot D

    
    Initial stiffness of p-y curve
        The initial stiffness :math:`k_{ini}` is infinite and can be capped from the Matlock original as in :py:func:`openpile.utils.py_curves.api_clay`:  


    p-y formulation (static loading, Neq = 1) 

        .. math::

            p = 
            \begin{cases} 
            \begin{split}
            0.5 \cdot P_{max} \left( \frac{y}{y_{50}} \right)^{0.33} & \text{  for } y \le 8 y_{50} \\ 
            \\
            P_{max} & \text{  for } y \gt 8 y_{50}
            \end{split}
            \end{cases}  

    p-y formulation (cyclic loading, Neq > 1)

        .. math::

            p = 
            \begin{cases} 
            \begin{split}
            0.5 \cdot P_{max} \left( \frac{y}{y_{50}} \right)^{0.33} & \text{  for } y \le 3 y_{50} \\ 
            \\
            0.72 \cdot P_{max} & \text{  for } y \gt 3 y_{50}
            \end{split}
            \end{cases}  

        For cyclic loading and curves above the transition zone ( i.e. :math:`X \le Xr`), 
        the p-y curve can be generated according to: 

        .. math::

            p = 
            \begin{cases} 
            \begin{split}
            0.5 \cdot P_{max} \left( \frac{y}{y_{50}} \right) & \text{  for } y \le 3 y_{50} \\ 
            \\
            0.72 \cdot P_{max} \left[ 1 - \left( 1 - \frac{X}{X_R} \right) \left( \frac{y - 3 y_{50}}{12 y_{50}} \right)  \right] & \text{  for } 3 y_{50} \lt y \le 15 y_{50} \\
            \\
            0.72 \cdot P_{max} \left( \frac{X}{X_R} \right) & \text{  for } y \gt 15 y_{50} \\
            \end{split}
            \end{cases}  
    """
    output_length = max(9, output_length)

    # important variables
    y50 = 2.5 * eps50 * D
    if X == 0.0 or Su == 0:
        Xr = 2.5 * D
    else:
        Xr = max((6 * D) / (sig / X * D / Su + J), 2.5 * D)

    # creation of 'y' array
    if ymax == 0.0:
        ymax = 16 * y50

    # Calculate Pmax (regular API)
    ## Pmax for shallow and deep zones (regular API)
    Pmax_shallow = (3 * Su + sig) * D + J * Su * X
    Pmax_deep = 9 * Su * D
    Pmax = min(Pmax_deep, Pmax_shallow)

    ylist_in = [0.0, 0.001 * y50, 0.01 * y50, 0.10 * y50, 1 * y50, 3 * y50, 8 * y50, 15 * y50, ymax]
    ylist_out = []
    for i in range(len(ylist_in)):
        if ylist_in[i] <= ymax:
            ylist_out.append(ylist_in[i])

    # determine y vector from 0 to ymax
    y = np.array(ylist_out, dtype=np.float32)
    add_values = output_length - len(y)
    add_y_values = []
    for _ in range(add_values):
        add_y_values.append(0.01 * y50 + random() * (8 - 0.01) * y50)
    y = np.append(y, add_y_values)
    y = np.sort(y)

    # define p vector
    p = np.zeros(shape=len(y), dtype=np.float32)

    for i in prange(len(y)):
        if kind == "static":
            # derive static curve
            if y[i] > 8 * y50:
                p[i] = Pmax
            else:
                p[i] = 0.5 * Pmax * (y[i] / y50) ** 0.33
        else:
            # derive cyclic curve
            if X <= Xr:
                if y[i] > 15 * y50:
                    p[i] = 0.7185 * Pmax * X / Xr
                elif y[i] > 3 * y50:
                    p[i] = 0.7185 * Pmax * (1 - (1 - X / Xr) * (y[i] - 3 * y50) / (12 * y50))
                else:
                    p[i] = 0.5 * Pmax * (y[i] / y50) ** 0.33

            elif X > Xr:
                if y[i] > 3 * y50:
                    p[i] = 0.7185 * Pmax
                else:
                    p[i] = 0.5 * Pmax * (y[i] / y50) ** 0.33

    return y, p


# API clay function
@njit(parallel=True, cache=True)
def modified_Matlock(
    sig: float,
    X: float,
    Su: float,
    eps50: float,
    D: float,
    J: float = 0.5,
    kind: str = "cyclic",
    ymax: float = 0.0,
    output_length: int = 20,
):
    """
    Creates the Modified Matlock for stiff clay p-y curve as defined in Bathacharya et al 2006 (see [BaCA06]).

    The modification takes places in the cyclic version of the curves. Static curves are kept the same as the original curves (see [Matl70]_), see :func:`Openpile.utils.py_curves.matlock_1970`.

    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    X: float
        Depth of the curve w.r.t. mudline [unit: m]
    Su : float
        Undrained shear strength [unit: kPa]
    eps50: float
        strain at 50% ultimate resistance [-]
    D: float
        Pile width [unit: m]
    J: float, by default 0.5
        empirical factor varying depending on clay stiffness
    kind: str, by default "cyclic"
        types of curves, can be of ("static","cyclic")
    ymax: float, by default 0.0
        maximum value of y, if null the maximum is calculated such that the whole curve is computed
    output_length: int, by default 20
        Number of discrete point along the springs, cannot be less than 8

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]


    See also
    --------
    :py:func:`openpile.utils.py_curves.api_clay`, :py:func:`openpile.utils.py_curves.matlock_1970`

    Notes
    -----

    Differences with standard Matlock (1970) formula
        For an undrained shear strength of 96 kPa (assumed as the threshold at which a clay is considered stiff),
        this formulation may be deemed more relevant to account for a more brittle fracture and degradation
        of the soil, see [BaCA06]_.

        .. figure:: _static/schematic_curves.png
            :width: 80%

            Schematic of original (soft clay response) and modified (stiff clay response), after [BaCA06]_.
    """

    output_length = max(8, output_length)

    # important variables
    y50 = 2.5 * eps50 * D
    if X == 0.0 or Su == 0:
        Xr = 2.5 * D
    else:
        Xr = max((6 * D) / (sig / X * D / Su + J), 2.5 * D)

    # creation of 'y' array
    if ymax == 0.0:
        ymax = 16 * y50

    # Calculate Pmax (regular API)
    ## Pmax for shallow and deep zones (regular API)
    Pmax_shallow = (3 * Su + sig) * D + J * Su * X
    Pmax_deep = 9 * Su * D
    Pmax = min(Pmax_deep, Pmax_shallow)

    ylist_in = [0.0, 0.01 * y50, 0.10 * y50, 1 * y50, 3 * y50, 8 * y50, 15 * y50, ymax]
    ylist_out = []
    for i in range(len(ylist_in)):
        if ylist_in[i] <= ymax:
            ylist_out.append(ylist_in[i])

    # determine y vector from 0 to ymax
    y = np.array(ylist_out, dtype=np.float32)
    add_values = output_length - len(y)
    add_y_values = []
    for _ in range(add_values):
        add_y_values.append(0.01 * y50 + random() * (8 - 0.01) * y50)
    y = np.append(y, add_y_values)
    y = np.sort(y)

    # define p vector
    p = np.zeros(shape=len(y), dtype=np.float32)

    for i in prange(len(y)):
        if kind == "static":
            # derive static curve
            if y[i] > 8 * y50:
                p[i] = Pmax
            else:
                p[i] = 0.5 * Pmax * (y[i] / y50) ** 0.33
        else:
            # derive cyclic curve
            if X <= Xr:
                if y[i] > 15 * y50:
                    p[i] = 0.5 * Pmax * X / Xr
                elif y[i] > 1 * y50:
                    p[i] = 0.5 * Pmax * (1 - (1 - X / Xr) * (y[i] - 1 * y50) / (14 * y50))
                else:
                    p[i] = 0.5 * Pmax * (y[i] / y50) ** 0.33

            elif X > Xr:
                if y[i] > 1 * y50:
                    p[i] = 0.5 * Pmax
                else:
                    p[i] = 0.5 * Pmax * (y[i] / y50) ** 0.33

    return y, p


@njit(parallel=True, cache=True)
def reese_weakrock(
    Ei: float,
    qu: float,
    RQD: float,
    xr: float,
    D: float,
    k: float = 0.0005,
    output_length=20,
):
    r"""creates the Reese weakrock p-y curve based on the work of Reese (1997), see reference [Rees97]_.

    Parameters
    ----------
    Ei : float
        initial rock mass modulus of rock [kPa]
    qu : float
        unconfined compressive strength of rock [kPa]
    RQD : float
        Rock Quality Designation [%]
    xr : float
        depth from rock surface [m]
    D : float
        pile width [m]
    k : float, optional
        dimensional constant, randing from 0.0005 to 0.00005, by default 0.0005
    output_length : int, optional
        length of output arrays, by default 20, cannot be less than 8

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]

        
    Notes
    -----
    This formulation was derived for weak rocks and based on [Rees97]_.
    This empirical model is mostly based on experimental data of pile load tests near San Francisco 
    where the rock unconfined compressive strength varies from 1.86 MPa near the surface to 16.0 MPa.
    Pressuremeter tests results were used by Reese in this formulation as the initial modulus of the rock.     

    The curve is characterized by a linear initial portion, a non-linear response for the remaining 
    part of the curve, and a maximum resistance value that can be mobilized

    Ultimate resistance of rock
        .. math::
            P_{max} = \alpha \cdot q_u \cdot D \left(1 + 1.4 \dfrac{x_r}{D}\right) \le 5.2 \alpha \cdot q_u \cdot D

        where:

        * :math:`\alpha = 1 - \dfrac{2}{3} \dfrac{\text{RQD}}{100}`
        * :math:`\text{RQD}` is the rock quality designation in percentage
        * :math:`q_u` is the unconfined compressive strength of rock
        * :math:`D` is the pile diameter
        * :math:`x_r` is the depth from rock surface

    Initial portion of p-y curve
        The initial part of the curve is defined for :math:`y \le yA`, with a linear p-y curve stiffness of :math:`E_{py_i}`.

        .. math::
            y_A &= \left(\dfrac{P_{max}}{2 \cdot y_{rm}^{0.25} \cdot E_{py_i}}\right)^{4/3} \\
            \\
            E_{py_i} &= \left(100 + 400 \dfrac{x_r}{3D} \right) E_i \le 500 E_i 


        where:

        * :math:`E_rm` is the rock mass modulus of rock
        * :math:`y_{rm} = k \cdot D`
        * :math:`k` is a dimensional constant ranging from 0.0005 to 0.00005

    Remaining non-linear response
        The remaining portion of the curve is defined with the following equation:

        .. math::
            p = \dfrac{P_{max}}{2} \left(\dfrac{y}{y_{rm}}\right)^{1/4} \le P_{max}
    """

    output_length = max(8, output_length)

    # Rqd forced to be within 0 and 100
    rqd = max(min(100, RQD), 0)

    # determine alpha
    alpha = 1.0 - 2 / 3 * rqd / 100

    # determine ultimate resistance of rock
    Pmax = min(5.2 * alpha * qu * D, alpha * qu * D * (1 + 1.4 * xr / D))

    # initial portion of p-y curve
    Epy_i = Ei * min(500, 100 + 400 * xr / (3 * D))

    # yA & yrm
    yrm = k * D
    yA = (Pmax / (2 * (yrm) ** 0.25 * Epy_i)) ** 1.333

    # define y
    ymax = max(1.05 * yA, (2 * yrm ** (0.25)) ** 4)
    y1 = np.linspace(yA, max(0.15 * ymax, 1.01 * yA), int(output_length / 2))
    y2 = np.linspace(max(yA * 1.02, 0.25 * ymax), ymax, output_length - len(y1) - 2)
    y = np.concatenate((np.array([0.0]), y1, y2, np.array([1.2 * ymax])))

    # define p
    p = np.zeros(y.size)
    for i in range(len(p)):
        if y[i] <= yA:
            p[i] = min(Pmax, Epy_i * y[i])
        else:
            p[i] = min(Pmax, Pmax / 2 * (y[i] / yrm) ** 0.25)

    return y, p


def custom_pisa_sand(
    sig: float,
    G0: float,
    D: float,
    X_ult: float,
    n: float,
    k: float,
    Y_ult: float,
    output_length: int = 20,
):
    """Creates a lateral spring with the PISA sand formulation and custom user inputs.

    Parameters
    ----------
    sig : float
        vertical/overburden effective stress [unit: kPa]
    G0 : float
        Small-strain shear modulus [unit: kPa]
    X_ult : float
        Normalized displacement at maximum strength
    k : float
        Normalized stiffness parameter
    n : float
        Normalized curvature parameter, must be between 0 and 1
    Y_ult : float
        Normalized maximum strength parameter
    output_length : int, optional
        Number of datapoints in the curve, by default 20

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]

    See also
    --------
    :py:func:`openpile.utils.py_curves.dunkirk_sand`, :py:func:`openpile.utils.hooks.dunkirk_sand_pisa_norm_param`

    Example
    -------

    .. plot::

        import matplotlib.pyplot as plt
        from openpile.utils.py_curves import custom_pisa_sand, dunkirk_sand


    """
    # calculate normsalised conic function
    y, p = conic(X_ult, n, k, Y_ult, output_length)

    # return non-normalised curve
    return y * (sig * D / G0), p * (sig * D)


@njit(cache=True)
def custom_pisa_clay(
    Su: float,
    G0: float,
    D: float,
    X_ult: float,
    n: float,
    k: float,
    Y_ult: float,
    output_length: int = 20,
):
    """
    Creates a spring from the PISA clay formulation
    published by Byrne et al 2020 (see [BHBG20]_) and calibrated based pile
    load tests at Cowden (north east coast of England).

    Parameters
    ----------
    Su : float
        Undrained shear strength [unit: kPa]
    G0 : float
        Small-strain shear modulus [unit: kPa]
    D : float
        Pile diameter [unit: m]
    X_ult : float
        Normalized displacement at maximum strength
    k : float
        Normalized stiffness parameter
    n : float
        Normalized curvature parameter, must be between 0 and 1
    Y_ult : float
        Normalized maximum strength parameter
    output_length : int, optional
        Number of datapoints in the curve, by default 20

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]

    See also
    --------
    :py:func:`openpile.utils.py_curves.cowden_clay`, :py:func:`openpile.utils.py_curves.bothkennar_clay`,
    :py:func:`openpile.utils.hooks.cowden_clay_pisa_norm_param`
    """
    # calculate normsalised conic function
    y, p = conic(X_ult, n, k, Y_ult, output_length)

    # return non-normalised curve
    return y * (Su * D / G0), p * (Su * D)
